import sys
import pprint
import click
import humanize
import pyaurorax
import textwrap
from texttable import Texttable

# globals
SUPPORTED_SOURCE_TYPES = [
    pyaurorax.sources.SOURCE_TYPE_EVENT_LIST,
    pyaurorax.sources.SOURCE_TYPE_GROUND,
    pyaurorax.sources.SOURCE_TYPE_HEO,
    pyaurorax.sources.SOURCE_TYPE_LEO,
    pyaurorax.sources.SOURCE_TYPE_LUNAR,
]
ALLOWED_FORMATS = [
    pyaurorax.FORMAT_BASIC_INFO,
    pyaurorax.FORMAT_BASIC_INFO_WITH_METADATA,
    pyaurorax.FORMAT_IDENTIFIER_ONLY,
    pyaurorax.FORMAT_FULL_RECORD
]


def __print_metadata_schema_table(ephemeris_schema=[], data_product_schema=[]):
    # init
    wrap_threshold = 40

    # set table lists
    table_schemas = []
    table_allowed_values = []
    table_data_types = []
    table_descriptions = []
    table_additional_descriptions = []
    table_field_names = []
    table_searchable = []
    for item in ephemeris_schema:
        table_schemas.append("ephemeris")
        table_field_names.append(item["field_name"])
        table_data_types.append(item["data_type"])
        table_allowed_values.append('\n'.join(textwrap.wrap(str(item["allowed_values"]), wrap_threshold)))
        table_descriptions.append('\n'.join(textwrap.wrap(item["description"], wrap_threshold)))
        if ("additional_description" in item):
            table_additional_descriptions.append('\n'.join(textwrap.wrap(item["additional_description"],
                                                                         wrap_threshold)))
        else:
            table_additional_descriptions.append("-")
        if ("searchable" in item):
            table_searchable.append(item["searchable"])
        else:
            table_searchable.append("-")
    for item in data_product_schema:
        table_schemas.append("data_product")
        table_field_names.append(item["field_name"])
        table_data_types.append(item["data_type"])
        table_allowed_values.append('\n'.join(textwrap.wrap(str(item["allowed_values"]), wrap_threshold)))
        table_descriptions.append('\n'.join(textwrap.wrap(item["description"], wrap_threshold)))
        if ("additional_description" in item):
            table_additional_descriptions.append('\n'.join(textwrap.wrap(item["additional_description"],
                                                                         wrap_threshold)))
        else:
            table_additional_descriptions.append("-")
        if ("searchable" in item):
            table_searchable.append(item["searchable"])
        else:
            table_searchable.append("-")

    # set header values
    table_headers = ["Schema", "Field Name", "Data Type",
                     "Allowed Values", "Searchable", "Description",
                     "Additional Description"]

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_field_names)):
        table.add_row([table_schemas[i],
                       table_field_names[i],
                       table_data_types[i],
                       table_allowed_values[i],
                       table_searchable[i],
                       table_descriptions[i],
                       table_additional_descriptions[i]])
    click.echo(table.draw())


def __print_single_data_source(ds, format):
    if (format == pyaurorax.FORMAT_IDENTIFIER_ONLY):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
    elif (format == pyaurorax.FORMAT_BASIC_INFO):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
    elif (format == pyaurorax.FORMAT_BASIC_INFO_WITH_METADATA):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
        if (ds.metadata == {}):
            click.echo("Metadata:\t\t%s" % (ds.metadata))
        else:
            click.echo("Metadata:\n%s" % (pprint.pformat(ds.metadata)))
    elif (format == pyaurorax.FORMAT_FULL_RECORD):
        click.echo("Identifier:\t\t%d" % (ds.identifier))
        click.echo("Program:\t\t%s" % (ds.program))
        click.echo("Platform:\t\t%s" % (ds.platform))
        click.echo("Instument Type:\t\t%s" % (ds.instrument_type))
        click.echo("Source Type:\t\t%s" % (ds.source_type))
        click.echo("Display Name:\t\t%s" % (ds.display_name))
        click.echo("Owner:\t\t\t%s" % (ds.owner))
        click.echo("Maintainers:\t\t%s" % (ds.maintainers))
        if (ds.metadata == {}):
            click.echo("Metadata:\t\t%s" % (ds.metadata))
        else:
            click.echo("Metadata:\n%s" % (pprint.pformat(ds.metadata)))
        if (ds.ephemeris_metadata_schema == []):
            click.echo("Ephemeris Schema:\t[]")
        else:
            click.echo("Ephemeris Schema:\tsee below table")
        if (ds.data_product_metadata_schema == []):
            click.echo("Data Product Schema:\t[]\n")
        else:
            click.echo("Data Product Schema:\tsee below table\n")
        __print_metadata_schema_table(ephemeris_schema=ds.ephemeris_metadata_schema,
                                      data_product_schema=ds.data_product_metadata_schema)


@click.group("sources", help="Interact with data sources")
def sources_group():
    pass


@sources_group.command("list", short_help="List data sources")
@click.option("--program", type=str, help="Filter using program")
@click.option("--platform", type=str, help="Filter using platform")
@click.option("--instrument-type", type=str, help="Filter using instrument type")
@click.option("--source-type", type=click.Choice(SUPPORTED_SOURCE_TYPES),
              help="Filter using source type")
@click.option("--owner", type=str, help="Filter using an owner")
@click.option("--order", type=click.Choice(["identifier", "program", "platform",
                                            "instrument_type", "display_name",
                                            "owner"]),
              default="identifier", show_default=True,
              help="Order results using a certain column")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def list(config, program, platform, instrument_type, source_type, owner, order, reversed_):
    """
    List data sources using the options to filter as desired
    """
    # get data sources
    try:
        sources = pyaurorax.sources.get_using_filters(program=program,
                                                      platform=platform,
                                                      instrument_type=instrument_type,
                                                      source_type=source_type,
                                                      owner=owner,
                                                      order=order)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # reverse
    if (reversed_ is True):
        sources = reversed(sources)

    # decide if we want to show the owner
    show_owner = False
    if (owner is not None or order == "owner"):
        show_owner = True

    # set table lists
    table_identifiers = []
    table_programs = []
    table_platforms = []
    table_instrument_types = []
    table_source_types = []
    table_display_names = []
    table_owners = []
    for source in sources:
        table_identifiers.append(source.identifier)
        table_programs.append(source.program)
        table_platforms.append(source.platform)
        table_instrument_types.append(source.instrument_type)
        table_source_types.append(source.source_type)
        table_display_names.append(source.display_name)
        table_owners.append(source.owner)

    # set header values
    table_headers = ["Identifier", "Display Name", "Program",
                     "Platform", "Instrument Type", "Source Type"]
    if (show_owner is True):
        table_headers.append("Owner")
    for i in range(0, len(table_headers)):
        if (table_headers[i].lower().replace(' ', '_') == order):
            table_headers[i] += " " + "\u2193"

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_identifiers)):
        if (show_owner is True):
            table.add_row([table_identifiers[i],
                           table_display_names[i],
                           table_programs[i],
                           table_platforms[i],
                           table_instrument_types[i],
                           table_source_types[i],
                           table_owners[i]])
        else:
            table.add_row([table_identifiers[i],
                           table_display_names[i],
                           table_programs[i],
                           table_platforms[i],
                           table_instrument_types[i],
                           table_source_types[i]])
    click.echo(table.draw())


@sources_group.command("search", short_help="Search for data sources")
@click.option("--programs", type=str,
              help="Search for program (comma separate for multiple values)")
@click.option("--platforms", type=str,
              help="Search for platform (comma separate for multiple values)")
@click.option("--instrument-types", type=str,
              help="Search for instrument type (comma separate for multiple values)")
@click.option("--order", type=click.Choice(["identifier", "program", "platform",
                                            "instrument_type", "display_name",
                                            "owner"]),
              default="identifier", show_default=True,
              help="Order results using a certain column")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def search(config, programs, platforms, instrument_types, order, reversed_):
    """
    Search for data sources using the options to filter as desired. Unlike
    the 'list' command filters, this command supports multiple programs,
    platforms, or instrument types (using commas).
    """
    # set programs values
    parsed_programs = []
    if (programs is None):
        pass
    elif (',' in programs):
        for p in programs.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_programs.append(p)
    else:
        parsed_programs = [programs]

    # set platforms values
    parsed_platforms = []
    if (platforms is None):
        pass
    elif (',' in platforms):
        for p in platforms.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_platforms.append(p)
    else:
        parsed_platforms = [platforms]

    # set instrument_types values
    parsed_instrument_types = []
    if (instrument_types is None):
        pass
    elif (',' in instrument_types):
        for p in instrument_types.split(','):
            p = p.strip()
            if (len(p) > 0):
                parsed_instrument_types.append(p)
    else:
        parsed_instrument_types = [instrument_types]

    # search for data sources
    try:
        sources = pyaurorax.sources.search(programs=parsed_programs,
                                           platforms=parsed_platforms,
                                           instrument_types=parsed_instrument_types,
                                           order=order)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # reverse
    if (reversed_ is True):
        sources = reversed(sources)

    # decide if we want to show the owner
    show_owner = False
    if (order == "owner"):
        show_owner = True

    # set table lists
    table_identifiers = []
    table_programs = []
    table_platforms = []
    table_instrument_types = []
    table_source_types = []
    table_display_names = []
    table_owners = []
    for source in sources:
        table_identifiers.append(source.identifier)
        table_programs.append(source.program)
        table_platforms.append(source.platform)
        table_instrument_types.append(source.instrument_type)
        table_source_types.append(source.source_type)
        table_display_names.append(source.display_name)
        table_owners.append(source.owner)

    # set header values
    table_headers = ["Identifier", "Display Name", "Program",
                     "Platform", "Instrument Type", "Source Type"]
    if (show_owner is True):
        table_headers.append("Owner")
    for i in range(0, len(table_headers)):
        if (table_headers[i].lower().replace(' ', '_') == order):
            table_headers[i] += " " + "\u2193"

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_identifiers)):
        if (show_owner is True):
            table.add_row([table_identifiers[i],
                           table_display_names[i],
                           table_programs[i],
                           table_platforms[i],
                           table_instrument_types[i],
                           table_source_types[i],
                           table_owners[i]])
        else:
            table.add_row([table_identifiers[i],
                           table_display_names[i],
                           table_programs[i],
                           table_platforms[i],
                           table_instrument_types[i],
                           table_source_types[i]])
    click.echo(table.draw())


@sources_group.command("get", short_help="Get a single data source")
@click.argument("program", type=str)
@click.argument("platform", type=str)
@click.argument("instrument_type", type=str)
@click.option("--format", type=click.Choice(ALLOWED_FORMATS),
              default=pyaurorax.FORMAT_BASIC_INFO,
              help="Amount of data about the data source to retrieve")
@click.pass_obj
def get(config, program, platform, instrument_type, format):
    """
    Get a single data source record

    \b
    PROGRAM           the program value
    PLATFORM          the platform value
    INSTRUMENT_TYPE   the instrument type value
    """
    # get the data source
    try:
        ds = pyaurorax.sources.get(program=program,
                                   platform=platform,
                                   instrument_type=instrument_type,
                                   format=format)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    __print_single_data_source(ds, format)


@sources_group.command("get_using_identifier",
                       short_help="Get a single data source (using an identifier)")
@click.argument("identifier", type=int)
@click.option("--format", type=click.Choice(ALLOWED_FORMATS),
              default=pyaurorax.FORMAT_BASIC_INFO,
              help="Amount of data about the data source to retrieve")
@click.pass_obj
def get_using_identifier(config, identifier, format):
    """
    Get a single data source record using an identifier

    \b
    IDENTIFIER     the identifier of the data source
    """
    # get data source
    try:
        ds = pyaurorax.sources.get_using_identifier(identifier, format=format)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    __print_single_data_source(ds, format)


@sources_group.command("get_stats", short_help="Get statistics about a data source")
@click.argument("identifier", type=int)
@click.option("--format", type=click.Choice(ALLOWED_FORMATS),
              default=pyaurorax.FORMAT_BASIC_INFO,
              help="Amount of data about the data source to retrieve")
@click.pass_obj
def get_stats(config, identifier, format):
    """
    Get statistics about a data source

    \b
    IDENTIFIER     the identifier of the data source
    """
    # get stats information
    try:
        stats = pyaurorax.sources.get_stats(identifier,
                                            format=format)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    click.echo("Data source:\t\t\t%s" % (stats.data_source))
    click.echo("Ephemeris count:\t\t%s" % (humanize.intcomma(stats.ephemeris_count)))
    click.echo("Earliest ephemeris loaded:\t%s" % (stats.earliest_ephemeris_loaded))
    click.echo("Latest ephemeris loaded:\t%s" % (stats.latest_ephemeris_loaded))
    click.echo("Data product count:\t\t%s" % (humanize.intcomma(stats.data_product_count)))
    click.echo("Earliest data product loaded:\t%s" % (stats.earliest_data_product_loaded))
    click.echo("Latest data product loaded:\t%s" % (stats.latest_data_product_loaded))


@sources_group.command("add", short_help="Add a data source")
@click.argument("program", type=str)
@click.argument("platform", type=str)
@click.argument("instrument_type", type=str)
@click.argument("source_type", type=click.Choice(SUPPORTED_SOURCE_TYPES))
@click.argument("display_name", type=str)
@click.option("--identifier", type=int, help="Custom identifier to use")
@click.pass_obj
def add(config, program, platform, instrument_type, source_type, display_name, identifier):
    """
    Add a data source

    \b
    PROGRAM           the program to set
    PLATFORM          the platform to set
    INSTRUMENT_TYPE   the instrument type to set
    SOURCE_TYPE       the source type to set
    DISPLAY_NAME      the display name to set
    """
    try:
        new_ds = pyaurorax.sources.DataSource(identifier=identifier,
                                              program=program,
                                              platform=platform,
                                              instrument_type=instrument_type,
                                              source_type=source_type,
                                              display_name=display_name)
        added_ds = pyaurorax.sources.add(new_ds)
        click.echo("Created data source successfully\n")
        click.echo(added_ds)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)


@sources_group.command("update", short_help="Update a data source")
@click.argument("identifier", type=str)
@click.option("--program", type=str, help="New program value")
@click.option("--platform", type=str, help="New platform value")
@click.option("--instrument-type", type=str, help="New instrument type value")
@click.option("--source-type", type=click.Choice(SUPPORTED_SOURCE_TYPES),
              help="New source type value")
@click.option("--display-name", type=str, help="New display name value")
@click.pass_obj
def update(config, identifier, program, platform, instrument_type, source_type, display_name):
    """
    Update a data source

    \b
    IDENTIFIER     the identifier of the data source
    """
    try:
        ds = pyaurorax.sources.update_partial(identifier,
                                              program=program,
                                              platform=platform,
                                              instrument_type=instrument_type,
                                              source_type=source_type,
                                              display_name=display_name)
        click.echo("Updated data source successfully\n")
        click.echo(ds)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)


@sources_group.command("delete", short_help="Delete a data source")
@click.argument("identifier", type=int)
@click.pass_obj
def delete(config, identifier):
    """
    Delete a data source

    \b
    IDENTIFIER     the identifier of the data source
    """
    try:
        pyaurorax.sources.delete(identifier)
        click.echo("Successfully deleted data source #%d" % (identifier))
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)



"""
Functions for using conjunction searches with Swarm-Aurora
"""

import webbrowser
import json
from typing import Dict, Union
from ...api import AuroraXRequest
from ...exceptions import AuroraXException
from ..classes.search import Search

# pdoc init
__pdoc__: Dict = {}


def get_url(search_obj: Search) -> str:
    """
    Get a URL that displays a conjunction search in the Swarm-Aurora
    Conjunction Finder

    Args:
        search_obj: a conjunction search object, must be a completed
                    search with the 'request_id' value populated

    Returns:
        the Swarm-Aurora Conjunction Finder URL for this conjunction search
    """
    return "https://swarm-aurora.com/conjunctionFinder/?aurorax_request_id=%s" % (search_obj.request_id)


def open_in_browser(search_obj: Search, browser: str = None) -> None:
    """
    In a browser, open a conjunction search in the Swarm-Aurora
    Conjunction Finder.

    Args:
        search_obj: a conjunction search object, must be a completed
                    search with the 'request_id' value populated
        browser: the browser type to load using. Default is your
                 default browser. Some common other options are
                 "google-chrome", "firefox", or "safari". For all available
                 options, refer to https://docs.python.org/3/library/webbrowser.html#webbrowser.get
    """
    url = get_url(search_obj)
    try:
        w = webbrowser.get(using=browser)
        w.open_new_tab(url)
    except Exception as e:
        if ("could not locate runnable browser" in str(e)):
            raise AuroraXException(("Error: selected browser '%s' not found, please try "
                                   "another. For the list of options, refer to "
                                    "https://docs.python.org/3/library/webbrowser.html#webbrowser.get") % (browser))


def create_custom_import_file(search_obj: Search,
                              filename: str = None,
                              returnDict: bool = False) -> Union[str, Dict]:
    """
    Generate a Swarm-Aurora custom import file for a given
    conjunction search

    Args:
        search_obj: a conjunction search object, must be a completed
                    search with the 'request_id' value populated
        filename: the output filename, default is 'swarmaurora_custom_import_file_{requestID}.json'
        returnDict: return the custom import file contents as a dictionary
                    instead of saving a file, default is False

    Returns:
        the filename of the saved custom import file, or a dictionary with the
        file contents if `returnDict` is set to True
    """
    # make request
    url = "https://swarm-aurora.com/conjunctionFinder/generate_custom_import_json?aurorax_request_id=%s" % (
        search_obj.request_id)
    req = AuroraXRequest(method="get",
                         url=url,
                         body=search_obj.query)
    res = req.execute()

    # return the contents as a dict if requested
    if (returnDict is True):
        return res.data

    # set default filename
    if (filename is None):
        filename = "swarmaurora_custom_import_%s.json" % (search_obj.request_id)

    # save data to file
    with open(filename, 'w', encoding='utf-8') as fp:
        json.dump(res.data, fp, indent=4)

    # return
    return filename



import click
import datetime
import pyaurorax
from dateutil.parser import parse


@click.group("util", help="Utility commands")
def utility_group():
    pass


@utility_group.command("ground_to_nbtrace",
                       short_help="Convert ground location to north B-trace location")
@click.argument("lat", type=click.FloatRange(min=-90.0, max=90.0))
@click.argument("lon", type=click.FloatRange(min=-180.0, max=180.0))
@click.option("--timestamp",
              type=str,
              default=datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
              show_default=True,
              help="timestamp to use when converting in ISO format "
              "(YYYY-mm-ddTHH:MM:SS), defaults to current time")
@click.pass_obj
def ground_to_nbtrace(config, lat, lon, timestamp):
    """
    Convert a geographic location (assumed to be on the ground) to a north
    B-trace location. Note that the timestamp is needed since the B-trace
    values are derived using magnetic coordinates

    \n
    Important! If you supply negative numbers, prefix it with '--' to
    tell the terminal to not evaluate dash (-) characters as options
    after it. For example, "aurorax-cli util ground_to_nbtrace -- 90.0 -120.0"

    \b
    LAT    the latitude, in geographic coordinates (-90 to 90)
    LON    the longitude, in geographic coordinates (-180 to 180)
    """
    # set location object
    input_location = pyaurorax.Location(lat=lat, lon=lon)
    click.echo("Inputted location:\t%s" % (input_location))

    # interpret time
    ts = parse(timestamp)
    click.echo("Timestamp:\t\t%s" % (ts.isoformat()))

    # get nbtrace location
    nbtrace = pyaurorax.util.ground_geo_to_nbtrace(input_location, ts)

    # output
    click.echo("\nDerived north B-trace:\t%s" % (nbtrace))


@utility_group.command("ground_to_sbtrace",
                       short_help="Convert ground location to south B-trace location")
@click.argument("lat", type=click.FloatRange(min=-90.0, max=90.0))
@click.argument("lon", type=click.FloatRange(min=-180.0, max=180.0))
@click.option("--timestamp",
              type=str,
              default=datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S"),
              show_default=True,
              help="timestamp to use when converting in ISO format "
              "(YYYY-mm-ddTHH:MM:SS), defaults to current time")
@click.pass_obj
def ground_to_sbtrace(config, lat, lon, timestamp):
    """
    Convert a geographic location (assumed to be on the ground) to a south
    B-trace location. Note that the timestamp is needed since the B-trace
    values are derived using magnetic coordinates

    \n
    Important! If you supply negative numbers, prefix it with '--' to
    tell the terminal to not evaluate dash (-) characters as options
    after it. For example, "aurorax-cli util ground_to_nbtrace -- 90.0 -120.0"

    \b
    LAT    the latitude, in geographic coordinates (-90 to 90)
    LON    the longitude, in geographic coordinates (-180 to 180)
    """
    # set location object
    input_location = pyaurorax.Location(lat=lat, lon=lon)
    click.echo("Inputted location:\t%s" % (input_location))

    # interpret time
    ts = parse(timestamp)
    click.echo("Timestamp:\t\t%s" % (ts.isoformat()))

    # get nbtrace location
    nbtrace = pyaurorax.util.ground_geo_to_sbtrace(input_location, ts)

    # output
    click.echo("\nDerived south B-trace:\t%s" % (nbtrace))



"""
Separted classes and functions used by the availability module.

Note that these classes and variables are all imported higher up at the
top of the availability module. They can be referenced from there instead
of digging in deeper to these submodules.
"""



"""
Class definition for a conjunction search
"""

import datetime
import itertools
from typing import Dict, List, Union, Optional
from .conjunction import Conjunction
from ...conjunctions import CONJUNCTION_TYPE_NBTRACE
from ...api import AuroraXRequest, AuroraXResponse, urls
from ...sources import DataSource, FORMAT_BASIC_INFO
from ...exceptions import AuroraXBadParametersException
from ...requests import (STANDARD_POLLING_SLEEP_TIME,
                         cancel as requests_cancel,
                         wait_for_data as requests_wait_for_data,
                         get_data as requests_get_data,
                         get_status as requests_get_status)

# pdoc init
__pdoc__: Dict = {}


class Search():
    """
    Class representing a conjunction search

    Attributes:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        distance: the maximum distance allowed between data sources when searching for
            conjunctions. This can either be a number (int or float), or a dictionary
            modified from the output of the "get_advanced_distances_combos()" function.
        ground: list of ground instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "calgary_apa_ml_v1",
                                "operator": "in",
                                "values": [ "classified as APA" ]
                            }
                        ]
                    }
                }]
        space: list of one or more space instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "nbtrace_region",
                                "operator": "in",
                                "values": [ "north auroral oval" ]
                            }
                        ]
                    },
                    "hemisphere": [
                        "northern"
                    ]
                }]
        events: list of one or more events search parameters, defaults to []

            Example:

                [{
                    "programs": [ "events" ],
                    "platforms": [ "toshi" ],
                    "instrument_types": [ "substorm onsets" ]
                }]
        conjunction_types: list of conjunction types, defaults to ["nbtrace"]. Options are
            in the pyaurorax.conjunctions module, or at the top level using the
            pyaurorax.CONJUNCTION_TYPE_* variables.
        epoch_search_precision: the time precision to which conjunctions are calculated. Can be
            30 or 60 seconds. Defaults to 60 seconds. Note - this parameter is under active
            development and still considered "alpha".
        response_format: JSON representation of desired data response format
        request: AuroraXResponse object returned when the search is executed
        request_id: unique ID assigned to the request by the AuroraX API
        request_url: unique URL assigned to the request by the AuroraX API
        executed: indicates if the search has been executed/started
        completed: indicates if the search has finished
        data_url: the URL where data is accessed
        query: the query for this request as JSON
        status: the status of the query
        data: the conjunctions found
        logs: all log messages outputed by the AuroraX API for this request

        Returns:
            a pyaurorax.conjunctions.Search object
    """

    def __init__(self, start: datetime.datetime,
                 end: datetime.datetime,
                 distance: Union[int, float, Dict[str, Union[int, float]]],
                 ground: Optional[List[Dict[str, str]]] = [],
                 space: Optional[List[Dict[str, str]]] = [],
                 events: Optional[List[Dict[str, str]]] = [],
                 conjunction_types: Optional[List[str]] = [CONJUNCTION_TYPE_NBTRACE],
                 epoch_search_precision: Optional[int] = 60,
                 response_format: Optional[Dict[str, bool]] = None):

        # set variables using passed in args
        self.start = start
        self.end = end
        self.ground = ground
        self.space = space
        self.events = events
        self.distance = distance
        self.conjunction_types = conjunction_types
        self.epoch_search_precision = epoch_search_precision
        self.response_format = response_format

        # initialize additional variables
        self.request: AuroraXResponse = None
        self.request_id: str = ""
        self.request_url: str = ""
        self.executed: bool = False
        self.completed: bool = False
        self.data_url: str = ""
        self.query: Dict = {}
        self.status: Dict = {}
        self.data: List[Union[Conjunction, Dict]] = []
        self.logs: List[Dict] = []

    def __str__(self):
        """
        String method

        Returns:
            string format of Conjunction Search object
        """
        return self.__repr__()

    def __repr__(self):
        """
        Object representation

        Returns:
            object representation of Conjunction Search object
        """
        return f"ConjunctionSearch(executed={self.executed}, " \
            f"completed={self.completed}, request_id='{self.request_id}')"

    def check_criteria_block_count_validity(self) -> None:
        """
        Check the number of of criteria blocks to see if there
        is too many. A max of 10 is allowed by the AuroraX
        conjunction search engine. An exception is raised if
        it was determined to have too many.

        Raises:
            pyaurorax.exceptions.AuroraXBadParametersException: too many criteria blocks are found
        """
        if ((len(self.ground) + len(self.space) + len(self.events)) > 10):
            raise AuroraXBadParametersException("Number of criteria blocks exceeds 10, "
                                                "please reduce the count")

    def get_advanced_distances_combos(self, default_distance: Union[int, float] = None) -> Dict:
        """
        Get the advanced distances combinations for this search

        Args:
            default_distance: the default distance to use, defaults to None

        Returns:
            the advanced distances combinations
        """
        # set input arrays
        options = []
        for i in range(0, len(self.ground)):
            options.append("ground%d" % (i + 1))
        for i in range(0, len(self.space)):
            options.append("space%d" % (i + 1))
        for i in range(0, len(self.events)):
            options.append("events%d" % (i + 1))

        # derive all combinations of options of size 2
        combinations = {}
        for element in itertools.combinations(options, r=2):
            combinations["%s-%s" % (element[0], element[1])] = default_distance

        # return
        return combinations

    def __fill_in_missing_distances(self, curr_distances: Dict) -> Dict:
        # get all distances possible
        all_distances = self.get_advanced_distances_combos()

        # go through current distances and fill in the values
        for curr_key, curr_value in curr_distances.items():
            curr_key_split = curr_key.split('-')
            curr_key1 = curr_key_split[0].strip()
            curr_key2 = curr_key_split[1].strip()
            for all_key in all_distances.keys():
                if (curr_key1 in all_key and curr_key2 in all_key):
                    # found the matching key, replace the value
                    all_distances[all_key] = curr_value

        # return
        return all_distances

    @property
    def distance(self) -> Union[int, float, Dict[str, Union[int, float]]]:
        """
        Property for the distance parameter

        Returns:
            the distance dictionary with all combinations
        """
        return self._distance

    @distance.setter
    def distance(self, distance: Union[int, float, Dict[str, Union[int, float]]]) -> None:
        # set distances to a dict if it's an int or float
        if (type(distance) is int or type(distance) is float):
            self._distance = self.get_advanced_distances_combos(default_distance=distance)  # type: ignore
        else:
            # is a dict, fill in any gaps
            self._distance = self.__fill_in_missing_distances(distance)  # type: ignore

    @property
    def query(self) -> Dict:
        """
        Property for the query value

        Returns:
            the query parameter
        """
        self._query = {
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
            "ground": self.ground,
            "space": self.space,
            "events": self.events,
            "conjunction_types": self.conjunction_types,
            "max_distances": self.distance,
            "epoch_search_precision": self.epoch_search_precision if self.epoch_search_precision in [30, 60] else 60,
        }
        return self._query

    @query.setter
    def query(self, query: Dict) -> None:
        self._query = query

    def execute(self) -> None:
        """
        Initiate a conjunction search request

        Raises:
            pyaurorax.exceptions.AuroraXBadParametersException: too many criteria blocks
        """
        # check number of criteria blocks
        self.check_criteria_block_count_validity()

        # do request
        url = urls.conjunction_search_url
        req = AuroraXRequest(method="post",
                             url=url,
                             body=self.query,
                             null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if res.status_code == 202:
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]

        # set request variable
        self.request = res

    def update_status(self, status: Optional[Dict] = None) -> None:
        """
        Update the status of this conjunction search request

        Args:
            status: the previously-retrieved status of this request (include
                to avoid requesting it from the API again), defaults to None
        """
        # get the status if it isn't passed in
        if (status is None):
            status = requests_get_status(self.request_url)

        # update request status by checking if data URI is set
        if (status["search_result"]["data_uri"] is not None):
            self.completed = True
            self.data_url = f'{urls.base_url}{status["search_result"]["data_uri"]}'

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this conjunction
        search request

        Returns:
            True if data is available, else False
        """
        self.update_status()
        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this conjunction search request
        """
        # check if request is completed
        if (self.completed is False):
            print("No data available, update status or check for data first")
            return

        # get data
        raw_data = requests_get_data(self.data_url, response_format=self.response_format)

        # set data variable
        if (self.response_format is not None):
            self.data = raw_data
        else:
            # cast data source objects
            for i in range(0, len(raw_data)):
                for j in range(0, len(raw_data[i]["data_sources"])):
                    ds = DataSource(**raw_data[i]["data_sources"][j], format=FORMAT_BASIC_INFO)
                    raw_data[i]["data_sources"][j] = ds

            # cast conjunctions
            self.data = [Conjunction(**c) for c in raw_data]

    def wait(self,
             poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
             verbose: Optional[bool] = False) -> None:
        """
        Block and wait until the request is complete and data is
        available for retrieval

        Args:
            poll_interval: time in seconds to wait between polling attempts, defaults
                to pyaurorax.requests.STANDARD_POLLING_SLEEP_TIME
            verbose: output poll times and other progress messages, defaults to False
        """
        url = urls.conjunction_request_url.format(self.request_id)
        self.update_status(requests_wait_for_data(url,
                                                  poll_interval=poll_interval,
                                                  verbose=verbose))

    def cancel(self,
               wait: Optional[bool] = False,
               poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
               verbose: Optional[bool] = False) -> int:
        """
        Cancel the conjunction search request

        This method returns immediately by default since the API processes
        this request asynchronously. If you would prefer to wait for it
        to be completed, set the 'wait' parameter to True. You can adjust
        the polling time using the 'poll_interval' parameter.

        Args:
            wait: wait until the cancellation request has been
                completed (may wait for several minutes)
            poll_interval: seconds to wait between polling
                calls, defaults to STANDARD_POLLING_SLEEP_TIME.
            verbose: output poll times and other progress messages, defaults
                to False

        Returns:
            1 on success

        Raises:
            pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
            pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
        """
        url = urls.conjunction_request_url.format(self.request_id)
        return requests_cancel(url, wait=wait, poll_interval=poll_interval, verbose=verbose)



"""
The data_products module is used to search and upload data
product records within AuroraX. One example of a data product
is a keogram.

Note that all functions and classes from submodules are all imported
at this level of the data_products module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# keogram data product type
DATA_PRODUCT_TYPE_KEOGRAM = "keogram"
"""
Data product type for keograms. Keograms are a 2-D
representation of a series of images, and are one of
the most popular data products that auroral science
uses. More information can be found at
https://docs.aurorax.space/about_the_data/standards/#keograms.
"""

# montage data product type
DATA_PRODUCT_TYPE_MONTAGE = "montage"
"""
Data product type for montages. Like keograms, montages are
another representation of a series of images. However, montages
are not a 2D representation but rather a collage of thumnbail
images for the period of time. An example can be found at
https://data.phys.ucalgary.ca/sort_by_project/THEMIS/asi/stream2/2021/12/28/gill_themis19/20211228__gill_themis19_full-montage.pgm.jpg
"""

# movie data product type
DATA_PRODUCT_TYPE_MOVIE = "movie"
"""
Data product type for movies. Movies are timelapse video
files of auroral data, usually as MP4 or MPEG. They can
consist of frames for a whole night, or an hour, and can
be at any cadence that is most appropriate.
"""

# summary plot data product type
DATA_PRODUCT_TYPE_SUMMARY_PLOT = "summary_plot"
"""
Data product type for summary plots. A summary plot can be any type
of plot that shows auroral data in a summary format, for example a
background-subtracted meridian scanning photometer plot showing
counts in Rayleighs.
"""

# data availability data product type
DATA_PRODUCT_TYPE_DATA_AVAILABILITY = "data_availability"
"""
Data product type for data availability. The AuroraX data availability
system does not account for times when data was not expected to be
collected, such as summer shutdowns due to inadequate night hours. This
data product type for 'data availbility' is meant to be used as a smarter
data availability mechanism for Aurora.
"""

# function and class imports
from .data_products import (search,
                            search_async,
                            upload,
                            delete_urls,
                            delete,
                            describe,
                            get_request_url)
from .classes.data_product import DataProduct
from .classes.search import Search

# pdoc imports and exports
from .data_products import __pdoc__ as __data_products_pdoc__
from .classes.data_product import __pdoc__ as __classes_data_product_pdoc__
from .classes.search import __pdoc__ as __classes_search_pdoc__
__pdoc__ = __data_products_pdoc__
__pdoc__ = dict(__pdoc__, **__classes_data_product_pdoc__)
__pdoc__ = dict(__pdoc__, **__classes_search_pdoc__)
__all__ = [
    "DATA_PRODUCT_TYPE_KEOGRAM",
    "DATA_PRODUCT_TYPE_MONTAGE",
    "DATA_PRODUCT_TYPE_MOVIE",
    "DATA_PRODUCT_TYPE_SUMMARY_PLOT",
    "DATA_PRODUCT_TYPE_DATA_AVAILABILITY",
    "search",
    "search_async",
    "upload",
    "delete_urls",
    "delete",
    "describe",
    "get_request_url",
    "DataProduct",
    "Search",
]



"""
Separted classes and functions used by the conjunctions module.

Note that these classes and variables are all imported higher up at the
top of the conjunctions module. They can be referenced from there instead
of digging in deeper to these submodules.
"""



"""
Separted classes and functions used by the api module.

Note that these classes and variables are all imported higher up at the
top of the api module. They can be referenced from there instead of digging
in deeper to these submodules.
"""



"""
Class definition for a statistics about a data source
"""

import datetime
from pydantic import BaseModel
from typing import Dict, Optional
from .data_source import DataSource

# pdoc init
__pdoc__: Dict = {}


class DataSourceStatistics(BaseModel):
    """
    Data source statistics object

    Attributes:
        data_source: the data source the statistics are associated with
        earliest_ephemeris_loaded: timestamp of the earliest ephemeris record
        latest_ephemeris_loaded: timestamp of the latest ephemeris record
        ephemeris_count: total number of ephemeris records for this data source
        earliest_data_product_loaded: timestamp of the earliest data_product record
        latest_data_product_loaded: timestamp of the latest data product record
        data_product_count: total number of ephemeris records for this data source
    """
    data_source: DataSource
    earliest_ephemeris_loaded: Optional[datetime.datetime] = None
    latest_ephemeris_loaded: Optional[datetime.datetime] = None
    ephemeris_count: int
    earliest_data_product_loaded: Optional[datetime.datetime] = None
    latest_data_product_loaded: Optional[datetime.datetime] = None
    data_product_count: int

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of DataSource object
        """
        return self.__repr__()



"""
Functions for interacting with metadata filters
"""

from typing import Dict, List, Optional
from ..sources import get_using_identifier, FORMAT_FULL_RECORD

# pdoc init
__pdoc__: Dict = {}


def validate(schema: List[Dict],
             record: Dict,
             quiet: Optional[bool] = False) -> bool:
    """
    Validate a metadata record against a schema. This checks that the
    key names match and there aren't fewer or more keys than expected.

    Args:
        schema: the metadata schema to validate against
        record: metadata record to validate

    Returns:
        True if the metadata record is valid, False if it is not
    """
    # set keys from schema and record
    schema_keys = sorted([i["field_name"] for i in schema])
    record_keys = sorted(record.keys())

    # look for bad keys
    if (schema_keys != record_keys):
        if (quiet is False):
            print("Metadata invalid, keys don't match")
        return False

    # found no bad keys
    return True


def get_ephemeris_schema(identifier: int) -> List[Dict]:
    """
    Retrieve the ephemeris metadata schema for a data source

    Args:
        identifier: the AuroraX data source ID

    Returns:
        the ephemeris metadata schema for the data source
    """
    # get the data source
    source_info = get_using_identifier(identifier, format=FORMAT_FULL_RECORD)

    # if there's an ephemeris metadata schema, return it
    if source_info.ephemeris_metadata_schema:
        return source_info.ephemeris_metadata_schema
    else:
        return []


def get_data_products_schema(identifier: int) -> List[Dict]:
    """
    Retrieve the data products metadata schema for a data source

    Args:
        identifier: the AuroraX data source ID

    Returns:
        the data products metadata schema for the data source
    """
    # get the data source
    source_info = get_using_identifier(identifier, format=FORMAT_FULL_RECORD)

    # if there's a data products metadata schema, return it
    if source_info.data_product_metadata_schema:
        return source_info.data_product_metadata_schema
    else:
        return []



"""
Class definition used for managing an API request
"""

import json
import requests
from pydantic import BaseModel
from typing import Optional, Dict, List, Union
from ..._internal.util import json_converter
from .response import AuroraXResponse
from ..api import get_api_key
from ...exceptions import (AuroraXMaxRetriesException,
                           AuroraXUnauthorizedException,
                           AuroraXNotFoundException,
                           AuroraXUnexpectedContentTypeException,
                           AuroraXUnexpectedEmptyResponse,
                           AuroraXException)

# pdoc init
__pdoc__: Dict = {}

# request globals
DEFAULT_RETRIES: int = 2
""" Number of retry attempts when requesting data from the API """

REQUEST_HEADERS: Dict = {
    "accept": "application/json",
    "Content-Type": "application/json"
}
""" The default headers sent as part of a request to the AuroraX API """

API_KEY_HEADER_NAME: str = "x-aurorax-api-key"
""" The API key header used when sending requests to the AuroraX API """


class AuroraXRequest(BaseModel):
    """
    AuroraX API request class

    Attributes:
        url: the URL to make the request against
        method: the HTTP method to use (get, post, put, delete, etc.)
        params: any URL parameters to send in the request, defaults to {}
        body: the body of the request (ie. post data), defaults to {}
        headers: any headers to send as part of the request (in addition to the default ones), default is {}
        null_response: signifies if we expect a response from the API that has no
            body/data in it (ie. requests to upload data that respond with just a
            202 status code), defaults to False
    """
    url: str
    method: str
    params: Optional[Dict] = {}
    body: Union[Optional[Dict], Optional[List]] = {}
    headers: Optional[Dict] = {}
    null_response: Optional[bool] = False

    def __merge_headers(self):
        # set initial headers
        all_headers = {
            "accept": "application/json",
            "Content-Type": "application/json"
        }

        # add headers passed into the class
        for key, value in self.headers.items():
            all_headers[key] = value

        # add api key
        api_key = get_api_key()
        if api_key:
            all_headers[API_KEY_HEADER_NAME] = api_key

        # return
        return all_headers

    def execute(self,
                limited_evaluation: Optional[bool] = False,
                skip_retry_logic: Optional[bool] = False) -> AuroraXResponse:
        """
        Execute an AuroraX request

        Args:
            limited_evaluation: don't evaluate the response after the retry
                mechanism, defaults to False
            skip_retry_logic: exclude the retry logic in the request, defaults
                to False

        Returns:
            an AuroraXResponse object

        Raises:
            pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
            pyaurorax.exceptions.AuroraXNotFoundException: requested resource was not found
            pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected content error
            pyaurorax.exceptions.AuroraXUnexpectedEmptyResponse: unexpected empty response
            pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
        """
        # sanitize data
        body_santized = json.dumps(self.body, default=json_converter)

        # make request
        req = requests.request(self.method,
                               self.url,
                               headers=self.__merge_headers(),
                               params=self.params,
                               data=body_santized)

        # retry request if needed
        if (skip_retry_logic is False):
            for i in range(0, DEFAULT_RETRIES):
                if (req.status_code == 500 and "text/plain" in req.headers["Content-Type"]):
                    if (i == (DEFAULT_RETRIES - 1)):
                        raise AuroraXMaxRetriesException("%s (%s)" % (req.content.decode(),
                                                                      req.status_code))
                    req = requests.request(self.method,
                                           self.url,
                                           headers=self.__merge_headers(),
                                           params=self.params,
                                           json=self.body,
                                           data=body_santized)
                else:
                    break

        # check if authorization worked (raised by API or by Nginx)
        if (req.status_code == 401):
            if (req.headers["Content-Type"] == "application/json"):
                if ("error_message" in req.json()):
                    # this will be an error message that the API meant to send
                    raise AuroraXUnauthorizedException("%s %s" % (req.status_code,
                                                                  req.json()["error_message"]))
                else:
                    raise AuroraXUnauthorizedException("Error 401: unauthorized")
            else:
                raise AuroraXUnauthorizedException("Error 401: unauthorized")

        # check for 404 error (raised by API or by Nginx)
        if (req.status_code == 404):
            if (req.headers["Content-Type"] == "application/json"):
                if ("error_message" in req.json()):
                    # this will be an error message that the API meant to send
                    raise AuroraXNotFoundException("%s %s" % (req.status_code,
                                                              req.json()["error_message"]))
                else:
                    # this will likely be a 404 from the java servlet
                    raise AuroraXNotFoundException("Error 404: not found")
            else:
                raise AuroraXNotFoundException("Error 404: not found")

        # check if we only want to do limited evaluation
        if (limited_evaluation is True):
            res = AuroraXResponse(request=req,
                                  data=None,
                                  status_code=req.status_code)
            return res

        # check content type
        if (self.null_response is False):
            if (req.headers["Content-Type"] == "application/json"):
                if (len(req.content) == 0):
                    raise AuroraXUnexpectedEmptyResponse("No response received")
                else:
                    response_data = req.json()
            else:
                raise AuroraXUnexpectedContentTypeException("%s (%s)" % (req.content.decode(),
                                                                         req.status_code))
        else:
            if (req.status_code in [200, 201, 202, 204]):
                response_data = None
            else:
                response_data = req.json()

        # check for server error
        if (req.status_code == 500):
            response_json = req.json()
            if ("error_message" in response_json):
                raise AuroraXException("%s (%s)" % (response_json["error_message"],
                                                    req.status_code))
            else:
                raise AuroraXException(response_json)

        # create reponse object
        res = AuroraXResponse(request=req,
                              data=response_data,
                              status_code=req.status_code)

        # return
        return res

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of AuroraXRequest
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of AuroraXRequest
        """
        return f"AuroraXRequest(method='{self.method}', url='{self.url}')"



"""
Functions for performing conjunction searches
"""

import datetime
import humanize
import warnings
from typing import Dict, List, Optional, Union
from .classes.search import Search
from ..api import AuroraXRequest, urls
from ..requests import STANDARD_POLLING_SLEEP_TIME

# pdoc init
__pdoc__: Dict = {}


def search(start: datetime.datetime,
           end: datetime.datetime,
           distance: Union[int, float, Dict[str, Union[int, float]]],
           ground: Optional[List[Dict[str, str]]] = [],
           space: Optional[List[Dict[str, str]]] = [],
           events: Optional[List[Dict[str, str]]] = [],
           conjunction_types: Optional[List[str]] = [],
           epoch_search_precision: Optional[int] = 60,
           response_format: Optional[Dict[str, bool]] = None,
           poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
           return_immediately: Optional[bool] = False,
           verbose: Optional[bool] = False) -> Search:
    """
    Search for conjunctions between data sources

    By default, this function will block and wait until the request completes and
    all data is downloaded. If you don't want to wait, set the 'return_immediately`
    value to True. The Search object will be returned right after the search has been
    started, and you can use the helper functions as part of that object to get the
    data when it's done.

    Args:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        distance: the maximum distance allowed between data sources when searching for
            conjunctions. This can either be a number (int or float), or a dictionary
            modified from the output of the "get_advanced_distances_combos()" function.
        ground: list of ground instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "calgary_apa_ml_v1",
                                "operator": "in",
                                "values": [ "classified as APA" ]
                            }
                        ]
                    }
                }]
        space: list of one or more space instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "nbtrace_region",
                                "operator": "in",
                                "values": [ "north auroral oval" ]
                            }
                        ]
                    },
                    "hemisphere": [
                        "northern"
                    ]
                }]
        events: list of one or more events search parameters, defaults to []

            Example:

                [{
                    "programs": [ "events" ],
                    "platforms": [ "toshi" ],
                    "instrument_types": [ "substorm onsets" ]
                }]
        conjunction_types: list of conjunction types, defaults to [] (meaning all conjunction
            types). Options are in the pyaurorax.conjunctions module, or at the top level using
            the pyaurorax.CONJUNCTION_TYPE_* variables.
        epoch_search_precision: the time precision to which conjunctions are calculated. Can be
            30 or 60 seconds. Defaults to 60 seconds. Note - this parameter is under active
            development and still considered "alpha".
        response_format: JSON representation of desired data response format
        poll_interval: seconds to wait between polling calls, defaults to
            pyaurorax.requests.STANDARD_POLLING_SLEEP_TIME
        return_immediately: initiate the search and return without waiting for data to
            be received, defaults to False
        verbose: show the progress of the request using the request log, defaults

    Returns:
        a pyaurorax.conjunctions.Search object
    """
    # create a Search object
    s = Search(start,
               end,
               distance,
               ground=ground,
               space=space,
               events=events,
               conjunction_types=conjunction_types,
               epoch_search_precision=epoch_search_precision,
               response_format=response_format)
    if (verbose is True):
        print(f"[{datetime.datetime.now()}] Search object created")

    # execute the search
    s.execute()
    if (verbose is True):
        print("[%s] Request submitted" % (datetime.datetime.now()))
        print("[%s] Request ID: %s" % (datetime.datetime.now(), s.request_id))
        print("[%s] Request details available at: %s" % (datetime.datetime.now(),
                                                         s.request_url))

    # return immediately if we wanted to
    if (return_immediately is True):
        return s

    # wait for data
    if (verbose is True):
        print("[%s] Waiting for data ..." % (datetime.datetime.now()))
    s.wait(poll_interval=poll_interval, verbose=verbose)

    # get the data
    if (verbose is True):
        print("[%s] Retrieving data ..." % (datetime.datetime.now()))
    s.get_data()

    # return response with the data
    if (verbose is True):
        print("[%s] Retrieved %s of data containing %d records" % (datetime.datetime.now(),
                                                                   humanize.filesize.naturalsize(
                                                                       s.status["search_result"]["file_size"]),
                                                                   s.status["search_result"]["result_count"]))

    # return
    return s


def search_async(start: datetime.datetime,
                 end: datetime.datetime,
                 distance: Union[int, float, Dict[str, Union[int, float]]],
                 ground: Optional[List[Dict[str, str]]] = [],
                 space: Optional[List[Dict[str, str]]] = [],
                 events: Optional[List[Dict[str, str]]] = [],
                 conjunction_types: Optional[List[str]] = [],
                 epoch_search_precision: Optional[int] = 60,
                 response_format: Optional[Dict[str, bool]] = None) -> Search:
    """
    Submit a request for a conjunctions search, return immediately

    The request will be done asynchronously by the API. Use the helper functions
    as part of the Search object returned to check for data and/or download it.
    If you don't want the search to return immediately and rather block until
    all data is downloaded, please use the 'search' function instead.

    .. deprecated::
        This function is deprecated as of 0.9.0. Please use the 'search' function
        with the 'return_immediately' flag set to True to get the same behaviour.
        This function will be removed in a future release.

    Args:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        distance: the maximum distance allowed between data sources when searching for
            conjunctions. This can either be a number (int or float), or a dictionary
            modified from the output of the "get_advanced_distances_combos()" function.
        ground: list of ground instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi"],
                    "platforms": ["gillam", "rabbit lake"],
                    "instrument_types": ["RGB"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "calgary_apa_ml_v1",
                                "operator": "in",
                                "values": [ "classified as APA" ]
                            }
                        ]
                    }
                }]
        space: list of one or more space instrument search parameters, defaults to []

            Example:

                [{
                    "programs": ["themis-asi", "swarm"],
                    "platforms": ["themisa", "swarma"],
                    "instrument_types": ["footprint"],
                    "ephemeris_metadata_filters": {
                        "logical_operator": "AND",
                        "expressions": [
                            {
                                "key": "nbtrace_region",
                                "operator": "in",
                                "values": [ "north auroral oval" ]
                            }
                        ]
                    },
                    "hemisphere": [
                        "northern"
                    ]
                }]
        events: list of one or more events search parameters, defaults to []

            Example:

                [{
                    "programs": [ "events" ],
                    "platforms": [ "toshi" ],
                    "instrument_types": [ "substorm onsets" ]
                }]
        conjunction_types: list of conjunction types, defaults to [] (meaning all conjunction
            types). Options are in the pyaurorax.conjunctions module, or at the top level using
            the pyaurorax.CONJUNCTION_TYPE_* variables.
        epoch_search_precision: the time precision to which conjunctions are calculated. Can be
            30 or 60 seconds. Defaults to 60 seconds. Note - this parameter is under active
            development and still considered "alpha".
        response_format: JSON representation of desired data response format

    Returns:
        a pyaurorax.conjunctions.Search object
    """
    warnings.warn("This function is deprecated and will be removed in a future release. Please "
                  "use the 'search' function with the 'return_immediately' flag to produce the "
                  "same behaviour.", DeprecationWarning, stacklevel=2)
    s = Search(start,
               end,
               distance,
               ground=ground,
               space=space,
               events=events,
               conjunction_types=conjunction_types,
               epoch_search_precision=epoch_search_precision,
               response_format=response_format)
    s.execute()
    return s


def describe(search_obj: Search) -> str:
    """
    Describe a conjunction search as an "SQL-like" string

    Args:
        search_obj: the conjunction search to describe

    Returns:
        the "SQL-like" string describing the conjunction search object
    """
    # make request
    req = AuroraXRequest(method="post",
                         url=urls.describe_conjunction_query,
                         body=search_obj.query)
    res = req.execute()

    # return
    return res.data


def get_request_url(request_id: str) -> str:
    """
    Get the conjunction search request URL for a given
    request ID. This URL can be used for subsequent
    pyaurorax.requests function calls. Primarily this method
    facilitates delving into details about a set of already-submitted
    conjunction searches.

    Args:
        request_id: the request identifier

    Returns:
        the request URL
    """
    url = urls.conjunction_request_url.format(request_id)
    return url



"""
Class definition for an ephemeris record
"""

import datetime
from pydantic import BaseModel
from typing import Dict, Optional
from ...location import Location
from ...sources import DataSource

# pdoc init
__pdoc__: Dict = {}


class Ephemeris(BaseModel):
    """
    Ephemeris object

    Attributes:
        data_source: data source that the ephemeris record is associated with
        epoch: timestamp for the record (assumed it is in UTC)
        location_geo: Location object containing geographic latitude and longitude
        location_gsm: Location object containing GSM latitude and longitude (leave
            empty for data sources with a type of 'ground')
        nbtrace: Location object with north B-trace geographic latitude and longitude
        sbtrace: Location object with south B-trace geographic latitude and longitude
        metadata: metadata for this record (arbitrary keys and values)
    """
    data_source: DataSource
    epoch: datetime.datetime
    location_geo: Location
    location_gsm: Optional[Location] = Location(lat=None, lon=None)
    nbtrace: Location
    sbtrace: Location
    metadata: Optional[Dict] = None

    def to_json_serializable(self) -> Dict:
        """
        Convert object to a JSON-serializable object (ie. translate
        datetime objects to strings)

        Returns:
            a dictionary object that is JSON-serializable
        """
        # init
        d = self.__dict__

        # format epoch as str
        if (type(d["epoch"]) is datetime.datetime):
            d["epoch"] = d["epoch"].strftime("%Y-%m-%dT%H:%M:00.000Z")

        # format location
        if (type(d["location_geo"]) is Location):
            d["location_geo"] = d["location_geo"].__dict__
        if (type(d["location_gsm"]) is Location):
            d["location_gsm"] = d["location_gsm"].__dict__
        if (type(d["nbtrace"]) is Location):
            d["nbtrace"] = d["nbtrace"].__dict__
        if (type(d["sbtrace"]) is Location):
            d["sbtrace"] = d["sbtrace"].__dict__

        # format metadata
        if (type(self.metadata) is dict):
            for key, value in self.metadata.items():
                if (type(value) is datetime.datetime or type(value) is datetime.date):
                    self.metadata[key] = self.metadata[key].strftime("%Y-%m-%dT%H:%M:%S.%f")
        if (type(self.metadata) is list):
            self.metadata = {}

        # format data source fields for query
        d["program"] = self.data_source.program
        d["platform"] = self.data_source.platform
        d["instrument_type"] = self.data_source.instrument_type
        del d["data_source"]

        # return
        return d

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of Ephemeris
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of Ephemeris
        """
        # shorten the metadata
        max_len = 20
        attr_metadata = f"{self.metadata}"
        if (len(attr_metadata) > max_len):
            attr_metadata = attr_metadata[0:max_len] + "...}"

        # return formatted representation
        return f"Ephemeris(data_source={repr(self.data_source)}, epoch={repr(self.epoch)}, " \
            f"location_geo={repr(self.location_geo)}, location_gsm={repr(self.location_gsm)}, " \
            f"nbtrace={repr(self.nbtrace)}, nbtrace={repr(self.nbtrace)}, " \
            f"metadata={attr_metadata})"



"""
AuroraX data sources are unique instruments that produce ephemeris or
data product records.

Note that all functions and classes from submodules are all imported
at this level of the sources module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# basic format
FORMAT_BASIC_INFO: str = "basic_info"
"""
Data sources are returned with the basic information: identifier,
program, platform, instrument type, source type, and display name
"""

# basic, but with metadata as well
FORMAT_BASIC_INFO_WITH_METADATA: str = "with_metadata"
"""
Data sources are returned with the basic information, plus the metadata
"""

# minimal, only the identifier
FORMAT_IDENTIFIER_ONLY: str = "identifier_only"
"""
Data sources are returned with only the identifier
"""

# full record, everything
FORMAT_FULL_RECORD: str = "full_record"
"""
Data sources are returned with all information about them. This
includes at least: identifier, program, platform, instrument type,
source type, display name, metadata, owner, maintainers, the
ephemeris metadata schema, and the data products meatadata schema.
"""

# default
FORMAT_DEFAULT: str = FORMAT_BASIC_INFO
"""
Default data source format (basic info)
"""

# ground source type
SOURCE_TYPE_GROUND: str = "ground"
"""
Data source 'source_type' category for a ground instrument
"""

# low-earth orbiting spacecraft source type
SOURCE_TYPE_LEO: str = "leo"
"""
Data source 'source_type' category for a low-earth orbiting satellite
"""

# highly-elliptical orbiting spacecraft source type
SOURCE_TYPE_HEO: str = "heo"
"""
Data source 'source_type' category for a highly-elliptical orbiting satellite
"""

# lunar orbiting spacecraft source type
SOURCE_TYPE_LUNAR: str = "lunar"
"""
Data source 'source_type' category for a lunar orbiting satellite
"""

# event list source type
SOURCE_TYPE_EVENT_LIST: str = "event_list"
"""
Data source 'source_type' category for a specially-curated event list
"""

# function and class imports
from .sources import (get,
                      get_using_filters,
                      get_using_identifier,
                      get_stats,
                      list,
                      search,
                      add,
                      delete,
                      update,
                      update_partial,
                      partial_update)
from .classes.data_source import DataSource
from .classes.data_source_stats import DataSourceStatistics

# pdoc imports and exports
from .sources import __pdoc__ as __sources_pdoc__
from .classes.data_source import __pdoc__ as __classes_data_source_pdoc__
from .classes.data_source_stats import __pdoc__ as __classes_data_source_stats_pdoc__
__pdoc__ = __sources_pdoc__
__pdoc__ = dict(__pdoc__, **__classes_data_source_pdoc__)
__pdoc__ = dict(__pdoc__, **__classes_data_source_stats_pdoc__)
__all__ = [
    "FORMAT_BASIC_INFO",
    "FORMAT_BASIC_INFO_WITH_METADATA",
    "FORMAT_FULL_RECORD",
    "FORMAT_IDENTIFIER_ONLY",
    "FORMAT_DEFAULT",
    "SOURCE_TYPE_EVENT_LIST",
    "SOURCE_TYPE_GROUND",
    "SOURCE_TYPE_HEO",
    "SOURCE_TYPE_LEO",
    "SOURCE_TYPE_LUNAR",
    "list",
    "search",
    "get",
    "get_using_filters",
    "get_using_identifier",
    "get_stats",
    "add",
    "delete",
    "update",
    "update_partial",
    "DataSource",
    "DataSourceStatistics",
]



"""
Class definition for a conjunction
"""

import datetime
from pydantic import BaseModel
from typing import Dict, List
from ...sources import DataSource

# pdoc init
__pdoc__: Dict = {}


class Conjunction(BaseModel):
    """
    Conjunction object

    Attributes:
        conjunction_type: the type of location data used when the
            conjunction was found (either be 'nbtrace' or 'sbtrace')
        start: start timestamp of the conjunction
        end: end timestamp of the conjunction
        data_sources: data sources in the conjunction
        min_distance: minimum kilometer distance of the conjunction
        max_distance: maximum kilometer distance of the conjunction
        events: the sub-conjunctions that make up this over-arching
            conjunction (the conjunctions between each set of two data
            sources)
    """
    conjunction_type: str
    start: datetime.datetime
    end: datetime.datetime
    data_sources: List[DataSource]
    min_distance: float
    max_distance: float
    events: List[Dict]

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of Conjunction object
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of Conjunction object
        """
        return f"Conjunction(start={repr(self.start)}, end={repr(self.end)}, " \
            f"min_distance={self.min_distance:.2f}, max_distance={self.max_distance:.2f}, " \
            "data_sources=[...], events=[...])"



"""
Class definition for a data product
"""

import datetime
from pydantic import BaseModel
from typing import Dict
from ...sources import DataSource

# pdoc init
__pdoc__: Dict = {}


class DataProduct(BaseModel):
    """
    Data product object

    Attributes:
        data_source: data source that the ephemeris record is associated with
        data_product_type: data product type ("keogram", "movie", "summary_plot")
        start: starting timestamp for the record (assumed it is in UTC), inclusive
        end: ending timestamp for the record (assumed it is in UTC), inclusive
        url: the URL of data product
        metdata: metadata for this record (arbitrary keys and values)
    """
    data_source: DataSource
    data_product_type: str
    start: datetime.datetime
    end: datetime.datetime
    url: str
    metadata: Dict

    def to_json_serializable(self) -> Dict:
        """
        Convert object to a JSON-serializable object (ie. translate
        datetime objects to strings)

        Returns:
            a dictionary object that is JSON-serializable
        """
        # init
        d = self.__dict__

        # format epoch as str
        if (type(d["start"]) is datetime.datetime):
            d["start"] = d["start"].strftime("%Y-%m-%dT%H:%M:00.000Z")
        if (type(d["end"]) is datetime.datetime):
            d["end"] = d["end"].strftime("%Y-%m-%dT%H:%M:00.000Z")

        # format metadata
        if (type(self.metadata) is dict):
            for key, value in self.metadata.items():
                if (type(value) is datetime.datetime or type(value) is datetime.date):
                    self.metadata[key] = self.metadata[key].strftime(
                        "%Y-%m-%dT%H:%M:%S.%f")
        if (type(self.metadata) is list):
            self.metadata = {}

        # format data source fields for query
        d["program"] = self.data_source.program
        d["platform"] = self.data_source.platform
        d["instrument_type"] = self.data_source.instrument_type
        del d["data_source"]

        # return
        return d

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of DataProduct object
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of DataProduct object
        """
        # shorten the metadata and url
        max_len = 20
        attr_metadata = f"{self.metadata}"
        if (len(attr_metadata) > max_len):
            attr_metadata = attr_metadata[0:max_len] + "...}"
        attr_url = f"{self.url}"
        if (len(attr_url) > max_len):
            attr_url = attr_url[0:max_len] + "..."

        # return formatted representation
        return f"DataProduct(data_source={repr(self.data_source)}, start={repr(self.start)}, " \
            f"end={repr(self.end)}, data_product_type='{self.data_product_type}', url='{attr_url}', " \
            f"metadata={attr_metadata})"



"""
Functions for interacting with data sources
"""

import warnings
from typing import List, Dict, Optional
from .classes.data_source import DataSource
from .classes.data_source_stats import DataSourceStatistics
from ..sources import FORMAT_FULL_RECORD
from ..api import AuroraXRequest, urls
from ..exceptions import (AuroraXNotFoundException,
                          AuroraXDuplicateException,
                          AuroraXBadParametersException,
                          AuroraXConflictException,
                          AuroraXException)

# pdoc init
__pdoc__: Dict = {}


def list(program: Optional[str] = None,
         platform: Optional[str] = None,
         instrument_type: Optional[str] = None,
         source_type: Optional[str] = None,
         owner: Optional[str] = None,
         format: Optional[str] = FORMAT_FULL_RECORD,
         order: Optional[str] = "identifier") -> List[DataSource]:
    """
    Retrieve all data source records (using params to filter as desired)

    Args:
        program: the program to filter for, defaults to None
        platform: the platform to filter for, defaults to None
        instrument_type: the instrument type to filter for, defaults to None
        source_type: the data source type to filter for, defaults to None.
            Options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.SOURCE_TYPE_* variables.
        owner: the owner's email address to filter for, defaults to None
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.
        order: the value to order results by (identifier, program, platform,
            instrument_type, display_name, owner), defaults to "identifier"

    Returns:
        any data sources matching the requested parameters

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
    """
    # make request
    params = {
        "program": program,
        "platform": platform,
        "instrument_type": instrument_type,
        "source_type": source_type,
        "owner": owner,
        "format": format,
    }
    req = AuroraXRequest(method="get",
                         url=urls.data_sources_url,
                         params=params)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    if len(res.data):
        return [DataSource(**ds, format=format) for ds in res.data]
    else:
        return []


def search(programs: Optional[List[str]] = [],
           platforms: Optional[List[str]] = [],
           instrument_types: Optional[List[str]] = [],
           format: Optional[str] = FORMAT_FULL_RECORD,
           order: Optional[str] = "identifier") -> List[DataSource]:
    """
    Search for data source records (using params to filter as desired)

    This is very similar to the 'list' function, however multiple programs,
    platforms, and/or instrument types can be supplied to this function. The
    'list' function only supports single values for those parameters.

    Args:
        programs: the programs to filter for, defaults to []
        platforms: the platforms to filter for, defaults to []
        instrument_type: the instrument types to filter for, defaults to []
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.
        order: the value to order results by (identifier, program, platform,
            instrument_type, display_name), defaults to "identifier"

    Returns:
        any data sources matching the requested parameters

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
    """
    # make request
    request_data = {
        "programs": programs,
        "platforms": platforms,
        "instrument_types": instrument_types,
    }
    params = {
        "format": format,
    }
    req = AuroraXRequest(method="post",
                         url=urls.data_sources_search_url,
                         params=params,
                         body=request_data)
    res = req.execute()

    # order results
    res.data = sorted(res.data, key=lambda x: x[order])

    # return
    if len(res.data):
        return [DataSource(**ds, format=format) for ds in res.data]
    else:
        return []


def get(program: str,
        platform: str,
        instrument_type: str,
        format: Optional[str] = FORMAT_FULL_RECORD) -> DataSource:
    """
    Retrieve a specific data source record

    Args:
        program: the program name
        platform: the platform name
        instrument_type: the instrument type name
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.

    Returns:
        the data source matching the requested parameters

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXNotFoundException: source not found
    """
    # get the data source
    data_sources = list(program=program,
                        platform=platform,
                        instrument_type=instrument_type,
                        format=format)

    # set results to the first thing found
    if (len(data_sources) == 1):
        return data_sources[0]
    elif (len(data_sources) > 1):
        warnings.warn("Found more than one data source matching this criteria, "
                      "returning the first (found %d)" % (len(data_sources)))
        return data_sources[0]
    else:
        raise AuroraXNotFoundException("No matching data source found")


def get_using_filters(program: Optional[str] = None,
                      platform: Optional[str] = None,
                      instrument_type: Optional[str] = None,
                      source_type: Optional[str] = None,
                      owner: Optional[str] = None,
                      format: Optional[str] = FORMAT_FULL_RECORD,
                      order: Optional[str] = "identifier") -> List[DataSource]:
    """
    Retrieve all data source records matching a filter

    Args:
        program: the program to filter for, defaults to None
        platform: the platform to filter for, defaults to None
        instrument_type: the instrument type to filter for, defaults to None
        source_type: the data source type to filter for, defaults to None.
            Options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.SOURCE_TYPE_* variables.
        owner: the owner's email address to filter for, defaults to None
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.
        order: the value to order results by (identifier, program, platform,
            instrument_type, display_name, owner), defaults to "identifier"

    Returns:
        any data sources matching the requested parameters

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
    """
    # get data sources
    data_sources = list(program=program,
                        platform=platform,
                        instrument_type=instrument_type,
                        source_type=source_type,
                        owner=owner,
                        format=format,
                        order=order)

    # return
    return data_sources


def get_using_identifier(identifier: int,
                         format: Optional[str] = FORMAT_FULL_RECORD) -> DataSource:
    """
    Retrieve data source record matching an identifier

    Args:
        identifier: the AuroraX unique ID for the data source
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.

    Returns:
        the data source matching the identifier

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
    """
    # make request
    params = {
        "format": format,
    }
    url = "%s/%d" % (urls.data_sources_url, identifier)
    req = AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # return
    return DataSource(**res.data, format=format)


def get_stats(identifier: int,
              format: Optional[str] = FORMAT_FULL_RECORD,
              slow: Optional[bool] = False) -> DataSourceStatistics:
    """
    Retrieve statistics for a data source

    Args:
        identifier: the AuroraX unique ID for the data source
        format: the format of the data sources returned, defaults to "full_record".
            Other options are in the pyaurorax.sources module, or at the top level
            using the pyaurorax.FORMAT_* variables.
        slow: retrieve the stats using a slower, but more accurate method, defaults to False

    Returns:
        information about the data source

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXNotFoundException: data source not found
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
    """
    # make request
    params = {
        "format": format,
        "slow": slow,
    }
    url = "%s/%d/stats" % (urls.data_sources_url, identifier)
    req = AuroraXRequest(method="get", url=url, params=params)
    res = req.execute()

    # cast data source record
    ds = DataSource(**res.data["data_source"], format=format)
    res.data["data_source"] = ds

    # return
    return DataSourceStatistics(**res.data)


def add(data_source: DataSource) -> DataSource:
    """
    Add a new data source to AuroraX

    Args:
        data_source: the data source to add (note: it must be a fully-defined
            DataSource object)

    Returns:
        the newly created data source

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXDuplicateException: duplicate data source, already exists
    """
    # set up request
    request_data = {
        "program": data_source.program,
        "platform": data_source.platform,
        "instrument_type": data_source.instrument_type,
        "source_type": data_source.source_type,
        "display_name": data_source.display_name,
        "ephemeris_metadata_schema": data_source.ephemeris_metadata_schema,
        "data_product_metadata_schema": data_source.data_product_metadata_schema,
        "metadata": data_source.metadata
    }
    if (data_source.identifier is not None):
        request_data["identifier"] = data_source.identifier

    # make request
    req = AuroraXRequest(method="post",
                         url=urls.data_sources_url,
                         body=request_data)
    res = req.execute()

    # evaluate response
    if (res.status_code == 409):
        raise AuroraXDuplicateException("%s - %s" % (res.data["error_code"],
                                                     res.data["error_message"]))

    # return
    try:
        return DataSource(**res.data)
    except Exception:
        raise AuroraXException("Could not create data source")


def delete(identifier: int) -> int:
    """
    Delete a data source from AuroraX

    Args:
        identifier: the AuroraX unique ID for the data source

    Returns:
        1 on success, raises error if an issue was encountered

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXNotFoundException: data source not found
        pyaurorax.exceptions.AuroraXConflictException: conflict of some type
    """
    # do request
    url = "%s/%d" % (urls.data_sources_url, identifier)
    req = AuroraXRequest(method="delete",
                         url=url,
                         null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):
        raise AuroraXBadParametersException("%s - %s" % (res.data["error_code"],
                                                         res.data["error_message"]))
    elif (res.status_code == 409):
        raise AuroraXConflictException("%s - %s" % (res.data["error_code"],
                                                    res.data["error_message"]))

    # return
    return 1


def update(data_source: DataSource) -> DataSource:
    """
    Update a data source in AuroraX

    This operation will fully replace the data source with the
    data_source argument passed in. Be sure that the data_source
    object is complete. If the data source is missing the value
    for identifier, program, platform, instrument type, source
    type, or display name, the update will fail and raise a
    AuroraXBadParametersException exception.

    Args:
        data_source: the data source to update (note: it must be a fully-defined
            DataSource object with the values set to what you want AuroraX to
            update it to)

    Returns:
        the updated data source

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXNotFoundException: data source not found
        pyaurorax.exceptions.AuroraXBadParametersException: missing parameters
    """
    # check to make sure the identifier, program, platform, instrument type,
    # source type, and display name are all set in the data source
    if not all([data_source.identifier, data_source.program, data_source.platform,
                data_source.instrument_type, data_source.source_type, data_source.display_name]):
        raise AuroraXBadParametersException("One or more required data source fields "
                                            "are missing, update operation aborted")

    # set URL
    url = f"{urls.data_sources_url}/{data_source.identifier}"

    # make request to update the data source passed in
    req_data = data_source.__dict__
    del req_data["format"]
    req = AuroraXRequest(method="put", url=url, body=req_data)
    req.execute()

    # return
    try:
        return get(data_source.program,
                   data_source.platform,
                   data_source.instrument_type,
                   format=FORMAT_FULL_RECORD)
    except Exception:
        raise AuroraXException("Could not update data source")


def update_partial(identifier: int,
                   program: Optional[str] = None,
                   platform: Optional[str] = None,
                   instrument_type: Optional[str] = None,
                   source_type: Optional[str] = None,
                   display_name: Optional[str] = None,
                   metadata: Optional[Dict] = None,
                   owner: Optional[str] = None,
                   maintainers: Optional[List[str]] = None,
                   ephemeris_metadata_schema: Optional[List[Dict]] = None,
                   data_product_metadata_schema: Optional[List[Dict]] = None) -> DataSource:
    """
    Partially update a data source in AuroraX (omitted fields are ignored)

    Args:
        identifier: the AuroraX unique ID for the data source, defaults to None
        program: the new program for the data source, defaults to None
        platform: the new platform for the data source, defaults to None
        instrument_type: the new instrument type for the data source, defaults to None
        source_type: the new source type for the data source, defaults to None. Options
            are in the pyaurorax.sources module, or at the top level using the
            pyaurorax.SOURCE_TYPE_* variables.
        display_name: the new display name for the data source, defaults to None
        metadata: the new metadata for the data source, defaults to None
        maintainers: the new maintainer AuroraX account email addresses, defaults to None
        ephemeris_metadata_schema: a list of dictionaries capturing the metadata
            keys and values that can appear in ephemeris records associated with
            the data source, defaults to None
        data_product_metadata_schema: a list of dictionaries capturing the metadata
            keys and values that can appear in data product records associated with
            the data source, defaults to None

    Returns:
        the updated data source

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXNotFoundException: data source not found
        pyaurorax.exceptions.AuroraXBadParametersException: missing parameters
    """
    # create a DataSource
    ds = DataSource(identifier=identifier,
                    program=program,
                    platform=platform,
                    instrument_type=instrument_type,
                    source_type=source_type,
                    display_name=display_name,
                    metadata=metadata,
                    owner=owner,
                    maintainers=maintainers,
                    ephemeris_metadata_schema=ephemeris_metadata_schema,
                    data_product_metadata_schema=data_product_metadata_schema)

    # set URL and request data
    url = f"{urls.data_sources_url}/{ds.identifier}"
    req_data = ds.__dict__
    del req_data["format"]

    # make request
    req = AuroraXRequest(method="patch", url=url, body=req_data)
    req.execute()

    # return
    try:
        return get_using_identifier(ds.identifier, format=FORMAT_FULL_RECORD)
    except Exception:
        raise AuroraXException("Could not update data source")


@DeprecationWarning
def partial_update(identifier: int,
                   program: Optional[str] = None,
                   platform: Optional[str] = None,
                   instrument_type: Optional[str] = None,
                   source_type: Optional[str] = None,
                   display_name: Optional[str] = None,
                   metadata: Optional[Dict] = None,
                   owner: Optional[str] = None,
                   maintainers: Optional[List[str]] = None,
                   ephemeris_metadata_schema: Optional[List[Dict]] = None,
                   data_product_metadata_schema: Optional[List[Dict]] = None) -> DataSource:
    warnings.warn("The 'partial_update' function has been deprecated as of 0.9.0, and will "
                  "be removed in a future release. Please use the 'update_partial' function "
                  "instead.", DeprecationWarning, stacklevel=2)
    return update_partial(identifier,
                          program=program,
                          platform=platform,
                          instrument_type=instrument_type,
                          source_type=source_type,
                          display_name=display_name,
                          metadata=metadata,
                          owner=owner,
                          maintainers=maintainers,
                          ephemeris_metadata_schema=ephemeris_metadata_schema,
                          data_product_metadata_schema=data_product_metadata_schema)



import requests
import click
import pyaurorax
from .availability import commands as availability_commands
from .conjunctions import commands as conjunctions_commands
from .data_products import commands as data_products_commands
from .ephemeris import commands as ephemeris_commands
from .sources import commands as sources_commands
from .util import commands as util_commands

# default context settings
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class Config(object):
    def __init__(self, verbose=False, api_key=None):
        self.verbose = verbose
        self.api_key = api_key


def __test_connectivity(quiet=False, return_json=False):
    r = requests.get(pyaurorax.api.get_base_url())
    if (r.status_code == 200):
        if (quiet is False):
            click.echo("Connectivity to the AuroraX API looks good!")
        if (return_json is True):
            return r.json()
    else:
        click.echo("Error connecting to AuroraX API, got a %d response" % (r.status_code))


@click.group(invoke_without_command=True)
@click.version_option(version="0.10.0")
@click.option("--api-key", type=str, help="Specify an API key")
@click.option("--api-base-url", type=str, help="Set the AuroraX API base URL")
@click.option("--verbose", "-v", is_flag=True, help="Enable verbose output")
@click.option("--test-connectivity", is_flag=True, help="Test connectivity to AuroraX API")
@click.pass_context
def cli(ctx, api_key, api_base_url, verbose, test_connectivity):
    """
    Welcome to the AuroraX CLI program!

    This program is meant to provide an easy interface with AuroraX
    from the command line. It uses the PyAuroraX library behind the scenes.
    """
    # set config
    ctx.obj = Config(verbose=verbose, api_key=api_key)

    # set the API base URL
    if (api_base_url is not None):
        pyaurorax.api.set_base_url(api_base_url)

    # authenticate
    if (api_key is not None):
        pyaurorax.authenticate(api_key)

    # evaluate options
    if (ctx.invoked_subcommand is None):
        if (test_connectivity is True):
            # evalualte --test-connectivity
            __test_connectivity(quiet=False)
        else:
            # no options called, output the help
            click.echo("""Welcome to the AuroraX CLI program!

This program is meant to provide an easy interface with AuroraX from the
command line. It uses the PyAuroraX library behind the scenes.

To learn more about usage, type:

  $> aurorax-cli --help
""")
    else:
        # subcommand was called, move on to that
        pass


# add sub commands
cli.add_command(availability_commands.availability_group)
cli.add_command(conjunctions_commands.conjunctions_group)
cli.add_command(data_products_commands.data_products_group)
cli.add_command(ephemeris_commands.ephemeris_group)
cli.add_command(sources_commands.sources_group)
cli.add_command(util_commands.utility_group)



import sys
import click
import datetime
import humanize
import pyaurorax
from typing import List
from texttable import Texttable


@click.group("availability", help="Retrieve data availability")
def availability_group():
    pass


def __print_availability_table(type: str,
                               availability: List[pyaurorax.availability.AvailabilityResult],
                               order: str) -> None:
    # check to see if there's anything to show
    if (len(availability) == 0):
        click.echo(f"No {type} availability information found!")
        return

    # set table lists
    table_identifiers = []
    table_programs = []
    table_platforms = []
    table_instrument_types = []
    table_source_types = []
    table_display_names = []
    table_dates = []
    table_counts = []
    for a in availability:
        # set which info to use, based on type
        if (type == "ephemeris"):
            dates_and_counts = a.available_ephemeris
        elif (type == "data products"):
            dates_and_counts = a.available_data_products
        else:
            click.echo("Unexpected error occurred, please open an issue on "
                       "the Github repository detailing how you were able "
                       "to make this message appear")

        # for each date
        for date, count in dates_and_counts.items():
            table_identifiers.append(a.data_source.identifier)
            table_programs.append(a.data_source.program)
            table_platforms.append(a.data_source.platform)
            table_instrument_types.append(a.data_source.instrument_type)
            table_source_types.append(a.data_source.source_type)
            table_display_names.append(a.data_source.display_name)
            table_dates.append(date)
            table_counts.append(humanize.intcomma(count))

    # set header values
    table_headers = ["Identifier", "Display Name", "Program",
                     "Platform", "Instrument Type", "Source Type",
                     "Date", "Available Records"]
    for i in range(0, len(table_headers)):
        if (table_headers[i].lower().replace(' ', '_') == order):
            table_headers[i] += " " + "\u2193"

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_identifiers)):
        table.add_row([table_identifiers[i],
                       table_display_names[i],
                       table_programs[i],
                       table_platforms[i],
                       table_instrument_types[i],
                       table_source_types[i],
                       table_dates[i],
                       table_counts[i]])
    click.echo(table.draw())


@availability_group.command("ephemeris",
                            short_help="Get ephemeris data availability info")
@click.argument("start_date", type=str)
@click.argument("end_date", type=str)
@click.option("--program", type=str, help="The program to filter for")
@click.option("--platform", type=str, help="The platform to filter for")
@click.option("--instrument-type", type=str, help="The instrument type to filter for")
@click.option("--source-type", type=str, help="The source type to filter for")
@click.option("--order", type=click.Choice(["identifier", "program", "platform",
                                            "instrument_type", "display_name"]),
              default="identifier", show_default=True,
              help="Order results using a certain column")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def ephemeris(config, start_date, end_date, program, platform, instrument_type, source_type,
              order, reversed_):
    """
    Get data availability information about ephemeris records

    \b
    START_DATE    the start date to retrieve info for, inclusive (YYYY/MM/DD)
    END_DATE      the end date to retrieve info for, inclusive (YYYY/MM/DD)
    """
    # set start and end datetime objects
    try:
        start_dt = datetime.datetime.strptime(start_date, "%Y/%m/%d")
    except Exception as e:
        click.echo("Error parsing start date, make sure it in YYYY/MM/DD format (%s)" % (str(e)))
        sys.exit(1)
    try:
        end_dt = datetime.datetime.strptime(end_date, "%Y/%m/%d")
    except Exception as e:
        click.echo("Error parsing end date, make sure it in YYYY/MM/DD format (%s)" % (str(e)))
        sys.exit(1)

    # get availability info
    try:
        a = pyaurorax.availability.ephemeris(start_dt,
                                             end_dt,
                                             program=program,
                                             platform=platform,
                                             instrument_type=instrument_type,
                                             source_type=source_type)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # order results
    if (order == "identifier"):
        a = sorted(a, key=lambda x: x.data_source.identifier)
    elif (order == "program"):
        a = sorted(a, key=lambda x: x.data_source.program)
    elif (order == "platform"):
        a = sorted(a, key=lambda x: x.data_source.platform)
    elif (order == "instrument_type"):
        a = sorted(a, key=lambda x: x.data_source.instrument_type)
    elif (order == "source_type"):
        a = sorted(a, key=lambda x: x.data_source.source_type)
    elif (order == "display_name"):
        a = sorted(a, key=lambda x: x.data_source.display_name)

    # reverse
    if (reversed_ is True):
        a = reversed(a)

    # print it out nicely
    __print_availability_table("ephemeris", a, order)


@availability_group.command("data_products",
                            short_help="Get data products data availability info")
@click.argument("start_date", type=str)
@click.argument("end_date", type=str)
@click.option("--program", type=str, help="The program to filter for")
@click.option("--platform", type=str, help="The platform to filter for")
@click.option("--instrument-type", type=str, help="The instrument type to filter for")
@click.option("--source-type", type=str, help="The source type to filter for")
@click.option("--order", type=click.Choice(["identifier", "program", "platform",
                                            "instrument_type", "display_name"]),
              default="identifier", show_default=True,
              help="Order results using a certain column")
@click.option("--reversed", "reversed_", is_flag=True, help="Reverse ordering")
@click.pass_obj
def data_products(config, start_date, end_date, program, platform, instrument_type, source_type,
                  order, reversed_):
    """
    Get data availability information about data product records

    \b
    START_DATE    the start date to retrieve info for, inclusive (YYYY/MM/DD)
    END_DATE      the end date to retrieve info for, inclusive (YYYY/MM/DD)
    """
    # set start and end datetime objects
    try:
        start_dt = datetime.datetime.strptime(start_date, "%Y/%m/%d")
    except Exception as e:
        click.echo("Error parsing start date, make sure it in YYYY/MM/DD format (%s)" % (str(e)))
        sys.exit(1)
    try:
        end_dt = datetime.datetime.strptime(end_date, "%Y/%m/%d")
    except Exception as e:
        click.echo("Error parsing end date, make sure it in YYYY/MM/DD format (%s)" % (str(e)))
        sys.exit(1)

    # get availability info
    try:
        a = pyaurorax.availability.data_products(start_dt,
                                                 end_dt,
                                                 program=program,
                                                 platform=platform,
                                                 instrument_type=instrument_type,
                                                 source_type=source_type)
    except pyaurorax.AuroraXException as e:
        click.echo("%s occurred: %s" % (type(e).__name__, e.args[0]))
        sys.exit(1)

    # print it out nicely
    __print_availability_table("data products", a, order)



"""
The requests module contains helper methods for retrieving data from
an AuroraX request.

Note that all functions and classes from submodules are all imported
at this level of the requests module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# function and class imports
from .requests import (FIRST_FOLLOWUP_SLEEP_TIME,
                       STANDARD_POLLING_SLEEP_TIME,
                       get_data,
                       get_logs,
                       get_status,
                       wait_for_data,
                       cancel)

# pdoc imports and exports
from .requests import __pdoc__ as __requests_pdoc__
__pdoc__ = __requests_pdoc__
__all__ = [
    "FIRST_FOLLOWUP_SLEEP_TIME",
    "STANDARD_POLLING_SLEEP_TIME",
    "get_data",
    "get_logs",
    "get_status",
    "wait_for_data",
    "cancel",
]



"""
Class definition for an ephemeris search
"""

import datetime
from typing import Dict, List, Union, Optional
from .ephemeris import Ephemeris
from ...api import AuroraXRequest, AuroraXResponse, urls
from ...sources import DataSource, FORMAT_BASIC_INFO
from ...exceptions import (AuroraXBadParametersException)
from ...requests import (STANDARD_POLLING_SLEEP_TIME,
                         cancel as requests_cancel,
                         wait_for_data as requests_wait_for_data,
                         get_data as requests_get_data,
                         get_status as requests_get_status)

# pdoc init
__pdoc__: Dict = {}


class Search():
    """
    Class representing an ephemeris search

    Note: At least one search criteria from programs, platforms, or instrument_types
    must be specified.

    Args:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        programs: list of programs to search through, defaults to None
        platforms: list of platforms to search through, defaults to None
        instrument_types: list of instrument types to search through, defaults to None
        metadata_filters: list of dictionaries describing metadata keys and
            values to filter on, defaults to None

            e.g. {
                "key": "string",
                "operator": "=",
                "values": [
                    "string"
                ]
            }
        metadata_filters_logical_operator: the logical operator to use when
            evaluating metadata filters (either 'AND' or 'OR'), defaults
            to "AND"
        response_format: JSON representation of desired data response format
        request: AuroraXResponse object returned when the search is executed
        request_id: unique ID assigned to the request by the AuroraX API
        request_url: unique URL assigned to the request by the AuroraX API
        executed: indicates if the search has been executed/started
        completed: indicates if the search has finished
        data_url: the URL where data is accessed
        query: the query for this request as JSON
        status: the status of the query
        data: the ephemeris records found
        logs: all log messages outputted by the AuroraX API for this request
    """

    def __init__(self,
                 start: datetime.datetime,
                 end: datetime.datetime,
                 programs: Optional[List[str]] = None,
                 platforms: Optional[List[str]] = None,
                 instrument_types: Optional[List[str]] = None,
                 metadata_filters: Optional[List[Dict]] = None,
                 metadata_filters_logical_operator: Optional[str] = "AND",
                 response_format: Optional[Dict] = None) -> None:

        # set variables using passed in args
        self.start = start
        self.end = end
        self.programs = programs
        self.platforms = platforms
        self.instrument_types = instrument_types
        self.metadata_filters = metadata_filters
        self.metadata_filters_logical_operator = metadata_filters_logical_operator
        self.response_format = response_format

        # initialize additional variables
        self.request: AuroraXResponse = None
        self.request_id: str = ""
        self.request_url: str = ""
        self.executed: bool = False
        self.completed: bool = False
        self.data_url: str = ""
        self.query: Dict = {}
        self.status: Dict = {}
        self.data: List[Union[Ephemeris, Dict]] = []
        self.logs: List[Dict] = []

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of Ephemeris Search object
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of Ephemeris Search object
        """
        return f"EphemerisSearch(executed={self.executed}, " \
            f"completed={self.completed}, request_id='{self.request_id}')"

    @property
    def query(self):
        """
        Property for the query value
        """
        self._query = {
            "data_sources": {
                "programs": [] if not self.programs else self.programs,
                "platforms": [] if not self.platforms else self.platforms,
                "instrument_types": [] if not self.instrument_types else self.instrument_types,
                "ephemeris_metadata_filters": {} if not self.metadata_filters
                else {
                    "logical_operator": self.metadata_filters_logical_operator,
                    "expressions": self.metadata_filters
                },
            },
            "start": self.start.strftime("%Y-%m-%dT%H:%M:%S"),
            "end": self.end.strftime("%Y-%m-%dT%H:%M:%S"),
        }
        return self._query

    @query.setter
    def query(self, query):
        self._query = query

    def execute(self) -> None:
        """
        Initiate ephemeris search request

        Raises:
            pyaurorax.exceptions.AuroraXBadParametersException: missing parameters
        """
        # check for at least one filter criteria
        if not (self.programs or self.platforms or self.instrument_types or self.metadata_filters):
            raise AuroraXBadParametersException("At least one filter criteria parameter "
                                                "besides 'start' and 'end' must be specified")

        # do request
        url = urls.ephemeris_search_url
        req = AuroraXRequest(method="post",
                             url=url,
                             body=self.query,
                             null_response=True)
        res = req.execute()

        # set request ID, request_url, executed
        self.executed = True
        if (res.status_code == 202):
            # request successfully dispatched
            self.executed = True
            self.request_url = res.request.headers["location"]
            self.request_id = self.request_url.rsplit("/", 1)[-1]

        # set the request variable
        self.request = res

    def update_status(self, status: Optional[Dict] = None) -> None:
        """
        Update the status of this ephemeris search request

        Args:
            status: the previously-retrieved status of this request (include
                to avoid requesting it from the API again), defaults to None
        """
        # get the status if it isn't passed in
        if (status is None):
            status = requests_get_status(self.request_url)

        # update request status by checking if data URI is set
        if (status["search_result"]["data_uri"] is not None):
            self.completed = True
            self.data_url = "%s%s" % (urls.base_url,
                                      status["search_result"]["data_uri"])

        # set class variable "status" and "logs"
        self.status = status
        self.logs = status["logs"]

    def check_for_data(self) -> bool:
        """
        Check to see if data is available for this ephemeris
        search request

        Returns:
            True if data is available, else False
        """
        self.update_status()
        return self.completed

    def get_data(self) -> None:
        """
        Retrieve the data available for this ephemeris search
        request
        """
        # check if completed yet
        if (self.completed is False):
            print("No data available, update status or check for data first")
            return

        # get data
        raw_data = requests_get_data(self.data_url, response_format=self.response_format)

        # set data variable
        if self.response_format is not None:
            self.data = raw_data
        else:
            # cast data source objects
            for i in range(0, len(raw_data)):
                ds = DataSource(**raw_data[i]["data_source"], format=FORMAT_BASIC_INFO)
                raw_data[i]["data_source"] = ds

            # cast ephemeris objects
            self.data = [Ephemeris(**e) for e in raw_data]

    def wait(self,
             poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
             verbose: Optional[bool] = False) -> None:
        """
        Block and wait for the request to complete and data is
        available for retrieval

        Args:
            poll_interval: time in seconds to wait between polling attempts,
                defaults to pyaurorax.requests.STANDARD_POLLING_SLEEP_TIME
            verbose: output poll times and other progress messages, defaults
                to False
        """
        url = urls.ephemeris_request_url.format(self.request_id)
        self.update_status(requests_wait_for_data(url,
                                                  poll_interval=poll_interval,
                                                  verbose=verbose))

    def cancel(self,
               wait: Optional[bool] = False,
               poll_interval: float = STANDARD_POLLING_SLEEP_TIME,
               verbose: Optional[bool] = False) -> int:
        """
        Cancel the ephemeris search request

        This method returns immediately by default since the API processes
        this request asynchronously. If you would prefer to wait for it
        to be completed, set the 'wait' parameter to True. You can adjust
        the polling time using the 'poll_interval' parameter.

        Args:
            wait: wait until the cancellation request has been
                completed (may wait for several minutes)
            poll_interval: seconds to wait between polling
                calls, defaults to STANDARD_POLLING_SLEEP_TIME.
            verbose: output poll times and other progress messages, defaults
                to False

        Returns:
            1 on success

        Raises:
            pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
            pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
        """
        url = urls.ephemeris_request_url.format(self.request_id)
        return requests_cancel(url, wait=wait, poll_interval=poll_interval, verbose=verbose)



"""
Functions for interacting with AuroraX requests
"""

import datetime
import time
from typing import Dict, List, Optional
from ..api.classes.request import AuroraXRequest
from ..exceptions import AuroraXDataRetrievalError
from ..location import Location

# pdoc init
__pdoc__: Dict = {}

# globals
FIRST_FOLLOWUP_SLEEP_TIME: float = 0.050  # 50ms
""" Initial sleep time when waiting for data """

STANDARD_POLLING_SLEEP_TIME: float = 1.0  # 1s
""" Polling sleep time when waiting for data (after the initial sleep time) """


def get_status(request_url: str) -> Dict:
    """
    Retrieve the status of a request

    Args:
        request_url: the URL of the request information

    Returns:
        the status information for the request
    """
    # do request
    req = AuroraXRequest(method="get", url=request_url)
    res = req.execute()

    # return
    return res.data


def get_data(data_url: str,
             response_format: Optional[Dict] = None,
             skip_serializing: Optional[bool] = False) -> List:
    """
    Retrieve the data for a request

    Args:
        data_url: the URL for the data of a request,
        response_format: the response format to send as post data, defaults
            to None
        skip_serializing: skip any object serializing, defaults to False

    Raises:
        pyaurorax.exceptions.AuroraXDataRetrievalError: error retrieving data

    Returns:
        the data for this request
    """
    # do request
    if (response_format is not None):
        req = AuroraXRequest(method="post",
                             url=data_url,
                             body=response_format)
    else:
        req = AuroraXRequest(method="get", url=data_url)
    res = req.execute()

    # check for error message
    if ("error" in res.data):
        raise AuroraXDataRetrievalError("%s: %s" % (res.data["error"]["error_code"],
                                                    res.data["error"]["error_message"]))

    # set data var
    data_result = res.data["result"]

    # serialize epochs and locations into datetimes and Locations
    if (skip_serializing is False):
        for i in range(0, len(data_result)):
            if ("epoch" in data_result[i]):
                data_result[i]["epoch"] = datetime.datetime.strptime(data_result[i]["epoch"],
                                                                     "%Y-%m-%dT%H:%M:%S")
            if ("location_geo" in data_result[i]):
                data_result[i]["location_geo"] = Location(lat=data_result[i]["location_geo"]["lat"],
                                                          lon=data_result[i]["location_geo"]["lon"])
            if ("location_gsm" in data_result[i]):
                data_result[i]["location_gsm"] = Location(lat=data_result[i]["location_gsm"]["lat"],
                                                          lon=data_result[i]["location_gsm"]["lon"])
            if ("nbtrace" in data_result[i]):
                data_result[i]["nbtrace"] = Location(lat=data_result[i]["nbtrace"]["lat"],
                                                     lon=data_result[i]["nbtrace"]["lon"])
            if ("sbtrace" in data_result[i]):
                data_result[i]["sbtrace"] = Location(lat=data_result[i]["sbtrace"]["lat"],
                                                     lon=data_result[i]["sbtrace"]["lon"])

    # return
    return data_result


def get_logs(request_url: str) -> List:
    """
    Retrieve the logs for a request

    Args:
        request_url: the URL of the request information

    Returns:
        the log messages for the request
    """
    # get status
    status = get_status(request_url)

    # return
    if ("logs" in status):
        return status["logs"]
    else:
        return []


def wait_for_data(request_url: str,
                  poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
                  verbose: Optional[bool] = False) -> Dict:
    """
    Block and wait for the data to be made available for a request

    Args:
        request_url: the URL of the request information
        poll_interval: seconds to wait between polling calls, defaults
            to STANDARD_POLLING_SLEEP_TIME
        verbose: output poll times and other progress messages, defaults to False

    Returns:
        the status information for the request
    """
    # get status
    status = get_status(request_url)

    # wait until request is done
    while (status["search_result"]["data_uri"] is None):
        time.sleep(poll_interval)
        if (verbose is True):
            print("[%s] Checking for data ..." % (datetime.datetime.now()))
        status = get_status(request_url)

    # return
    if (verbose is True):
        print("[%s] Data is now available" % (datetime.datetime.now()))
    return status


def cancel(request_url: str,
           wait: Optional[bool] = False,
           poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
           verbose: Optional[bool] = False) -> int:
    """
    Cancel the request at the given URL.

    This method returns immediately by default since the API processes
    this request asynchronously. If you would prefer to wait for it
    to be completed, set the 'wait' parameter to True. You can adjust
    the polling time using the 'poll_interval' parameter.

    Args:
        request_url: the URL string of the request to be canceled
        wait: set to True to block until the cancellation request
            has been completed (may wait for several minutes)
        poll_interval: seconds to wait between polling
            calls, defaults to STANDARD_POLLING_SLEEP_TIME.
        verbose: if True then output poll times and other
            progress, defaults to False

    Returns:
        1 on success

    Raises:
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
    """
    # do request
    req = AuroraXRequest(method="delete",
                         url=request_url,
                         null_response=True)
    req.execute()

    # return immediately if we don't want to wait
    if (wait is False):
        return 1

    # get status
    status = get_status(request_url)

    # wait for request to be cancelled
    while (status["search_result"]["data_uri"] is None and status["search_result"]["error_condition"] is False):
        time.sleep(poll_interval)
        if (verbose is True):
            print("[%s] Checking for cancellation status ..." % (datetime.datetime.now()))
        status = get_status(request_url)

    # return
    if (verbose is True):
        print("[%s] The request has been cancelled" % (datetime.datetime.now()))
    return 1



"""
This module is the under-the-hood interface for RESTful API
requests. It provides helper functions that the PyAuroraX library
uses to make robust requests.

Note that all functions and classes from submodules are all imported
at this level of the api module. They can be referenced from here
instead of digging in deeper to the submodules.
"""

# endpoint URLs
DEFAULT_BASE_URL: str = "https://api.aurorax.space"
""" The default API base URL to use when sending requests """

# function and class imports
from .classes.request import (DEFAULT_RETRIES,
                              REQUEST_HEADERS,
                              API_KEY_HEADER_NAME,
                              AuroraXRequest)
from .classes.response import AuroraXResponse
from .api import (urls,
                  get_api_key,
                  authenticate,
                  set_base_url,
                  get_base_url,
                  reset_base_url)

# pdoc import and exports
from .api import __pdoc__ as __api_pdoc__
from .classes.request import __pdoc__ as __classes_request_pdoc__
from .classes.response import __pdoc__ as __classes_response_pdoc__
__pdoc__ = __api_pdoc__
__pdoc__ = dict(__pdoc__, **__classes_request_pdoc__)
__pdoc__ = dict(__pdoc__, **__classes_response_pdoc__)
__all__ = [
    "DEFAULT_BASE_URL",
    "DEFAULT_RETRIES",
    "REQUEST_HEADERS",
    "API_KEY_HEADER_NAME",
    "AuroraXRequest",
    "AuroraXResponse",
    "urls",
    "get_api_key",
    "authenticate",
    "set_base_url",
    "get_base_url",
    "reset_base_url",
]



"""
Helper functions for calculating the north and south B-trace geographic
locations for ground-based instruments.
"""

import datetime
import warnings
from typing import Dict
from ..location import Location

# import aacgmv2 if installed
try:
    import aacgmv2
    __aacgm_found = True
except ModuleNotFoundError:
    __aacgm_found = False

# pdoc init
__pdoc__: Dict = {}


def __calculate_btrace(geo_location: Location, dt: datetime.datetime) -> Location:
    # convert to magnetic coordinates
    mag_location = aacgmv2.convert_latlon(geo_location.lat,
                                          geo_location.lon,
                                          0.0,
                                          dt,
                                          method_code="G2A")

    # change magnetic latitude to other hemisphere
    mag_location = (mag_location[0] * -1.0,
                    mag_location[1],
                    mag_location[2])

    # convert magnetic coordinates back to geographic
    btrace_aacgm = aacgmv2.convert_latlon(mag_location[0],
                                          mag_location[1],
                                          mag_location[2],
                                          dt,
                                          method_code="A2G")

    # return as Location object
    return Location(lat=btrace_aacgm[0], lon=btrace_aacgm[1])


def ground_geo_to_nbtrace(geo_location: Location,
                          timestamp: datetime.datetime) -> Location:
    """
    Convert geographic location to North B-Trace geographic
    location

    The timestamp is required because when calculating the B-trace
    values, the location is converted into geomagnetic coordinates.
    This conversion is different based on the timestamp since the
    magnetic coordinates change over time.

    Note: aacgmv2 must be installed. To install it, you can run
    "python -m pip install pyaurorax[aacgmv2]".

    Args:
        geo_location: a Location object representing the
            geographic location
        dt: timestamp for this set of lat and lons

    Returns:
        the north B-trace location as a Location object
    """
    # check to make sure aacgmv2 is installed
    if (__aacgm_found is False):
        warnings.warn("The aacgmv2 package is not installed, so an unchanged "
                      "location object will be returned. For this function to "
                      "work, please install it using 'pip install pyaurorax[aacgmv2]'.")
        return geo_location

    # check if location is in northern hemisphere
    if (geo_location.lat is not None and geo_location.lat >= 0.0):
        # northern hemisphere, north b-trace is the same as geographic location
        return geo_location

    # calculate South B-trace and return
    sbtrace = __calculate_btrace(geo_location, timestamp)
    return sbtrace


def ground_geo_to_sbtrace(geo_location: Location,
                          timestamp: datetime.datetime) -> Location:
    """
    Convert geographic location to South B-Trace geographic
    location

    The timestamp is required because when calculating the B-trace
    values, the location is converted into geomagnetic coordinates.
    This conversion is different based on the timestamp since the
    magnetic coordinates change over time.

    Note: aacgmv2 must be installed. To install it, you can run
    "python -m pip install pyaurorax[aacgmv2]".

    Args:
        geo_location: a Location object representing the
            geographic location
        dt: timestamp for this set of lat and lons

    Returns:
        the south B-trace location as a Location object
    """
    # check to make sure aacgmv2 is installed
    if (__aacgm_found is False):
        warnings.warn("The aacgmv2 package is not installed, so an unchanged "
                      "location object will be returned. For this function to "
                      "work, please install it using 'pip install pyaurorax[aacgmv2]'.")
        return geo_location

    # check if location is in southern hemisphere
    if (geo_location.lat is not None and geo_location.lat < 0.0):
        # southern hemisphere, south b-trace is the same as geographic location
        return geo_location

    # calculate North B-trace and return
    nbtrace = __calculate_btrace(geo_location, timestamp)
    return nbtrace



"""
The ephemeris module is used to search and upload ephemeris records
within AuroraX.

Note that all functions and classes from submodules are all imported
at this level of the ephemeris module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# function and class imports
from .ephemeris import (search_async,
                        search,
                        upload,
                        delete,
                        describe,
                        get_request_url)
from .classes.ephemeris import Ephemeris
from .classes.search import Search

# pdoc imports and exports
from .ephemeris import __pdoc__ as __ephemeris_pdoc__
from .classes.ephemeris import __pdoc__ as __classes_ephemeris_pdoc__
from .classes.search import __pdoc__ as __classes_search_pdoc__
__pdoc__ = __ephemeris_pdoc__
__pdoc__ = dict(__pdoc__, **__classes_ephemeris_pdoc__)
__pdoc__ = dict(__pdoc__, **__classes_search_pdoc__)
__all__ = [
    "search_async",
    "search",
    "upload",
    "delete",
    "describe",
    "get_request_url",
    "Ephemeris",
    "Search",
]



"""
Functions for performing data product searches
"""

import datetime
import humanize
import warnings
from typing import Dict, List, Optional
from .classes.data_product import DataProduct
from .classes.search import Search
from ..sources import (DataSource,
                       list as sources_list)
from ..exceptions import (AuroraXValidationException,
                          AuroraXUploadException,
                          AuroraXBadParametersException,
                          AuroraXException)
from ..requests import STANDARD_POLLING_SLEEP_TIME
from ..api import (AuroraXRequest,
                   urls as api_urls)

# pdoc init
__pdoc__: Dict = {}


def __validate_data_source(identifier: int,
                           records: List[DataProduct]) -> Optional[DataProduct]:
    # get all current sources
    sources = {source.identifier: source for source in sources_list()}
    if identifier not in sources.keys():
        raise AuroraXValidationException(f"Data source with unique identifier "
                                         "{identifier} could not be found")

    # process each record to make sure the program/platform/instrument_type matches
    # the identifier found for the data source
    for record in records:
        # check the identifier, program name, platform name, and instrument type
        try:
            reference = sources[record.data_source.identifier]
        except KeyError:
            raise AuroraXValidationException(f"Data source with unique identifier "
                                             "{record.data_source.identifier} could "
                                             "not be found")

        # check if it's a bad record
        if not (record.data_source.program == reference.program
                and record.data_source.platform == reference.platform
                and record.data_source.instrument_type == reference.instrument_type):
            return record

    # found no bad records
    return None


def search(start: datetime.datetime,
           end: datetime.datetime,
           programs: Optional[List[str]] = None,
           platforms: Optional[List[str]] = None,
           instrument_types: Optional[List[str]] = None,
           data_product_types: Optional[List[str]] = None,
           metadata_filters: Optional[List[Dict]] = None,
           metadata_filters_logical_operator: Optional[str] = None,
           response_format: Optional[Dict] = None,
           poll_interval: Optional[float] = STANDARD_POLLING_SLEEP_TIME,
           return_immediately: Optional[bool] = False,
           verbose: Optional[bool] = False) -> Search:
    """
    Search for data product records

    By default, this function will block and wait until the request completes and
    all data is downloaded. If you don't want to wait, set the 'return_immediately`
    value to True. The Search object will be returned right after the search has been
    started, and you can use the helper functions as part of that object to get the
    data when it's done.

    Note: At least one search criteria from programs, platforms, or
    instrument_types, must be specified.

    Args:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        programs: list of programs to search through, defaults to None
        platforms: list of platforms to search through, defaults to None
        instrument_types: list of instrument types to search through, defaults to None
        data_product_types: list of dictionaries describing data product
            types to filter on e.g. "keogram", defaults to None. Options are in the
            pyaurorax.data_products module, or at the top level using the
            pyaurorax.DATA_PRODUCT_TYPE* variables.
        metadata_filters: list of dictionaries describing metadata keys and
            values to filter on, defaults to None

            Example:

                [{
                    "key": "nbtrace_region",
                    "operator": "in",
                    "values": ["north polar cap"]
                }]
        metadata_filters_logical_operator: the logical operator to use when
            evaluating metadata filters (either 'AND' or 'OR'), defaults
            to "AND"
        response_format: JSON representation of desired data response format
        poll_interval: time in seconds to wait between polling attempts, defaults
            to pyaurorax.requests.STANDARD_POLLING_SLEEP_TIME
        return_immediately: initiate the search and return without waiting for data to
            be received, defaults to False
        verbose: output poll times and other progress messages, defaults to False

    Returns:
        a pyaurorax.data_products.Search object
    """
    # create a Search() object
    s = Search(start,
               end,
               programs=programs,
               platforms=platforms,
               instrument_types=instrument_types,
               data_product_types=data_product_types,
               metadata_filters=metadata_filters,
               metadata_filters_logical_operator=metadata_filters_logical_operator,
               response_format=response_format)
    if (verbose is True):
        print("[%s] Search object created" % (datetime.datetime.now()))

    # execute the search
    s.execute()
    if (verbose is True):
        print("[%s] Request submitted" % (datetime.datetime.now()))
        print("[%s] Request ID: %s" % (datetime.datetime.now(), s.request_id))
        print("[%s] Request details available at: %s" % (datetime.datetime.now(),
                                                         s.request_url))

    # return immediately if we wanted to
    if (return_immediately is True):
        return s

    # wait for data
    if (verbose is True):
        print("[%s] Waiting for data ..." % (datetime.datetime.now()))
    s.wait(poll_interval=poll_interval, verbose=verbose)

    # get the data
    if (verbose is True):
        print("[%s] Retrieving data ..." % (datetime.datetime.now()))
    s.get_data()

    # return response with the data
    if (verbose is True):
        print("[%s] Retrieved %s of data containing %d records" % (datetime.datetime.now(),
                                                                   humanize.filesize.naturalsize(
                                                                       s.status["search_result"]["file_size"]),
                                                                   s.status["search_result"]["result_count"]))
    return s


def search_async(start: datetime.datetime,
                 end: datetime.datetime,
                 programs: Optional[List[str]] = None,
                 platforms: Optional[List[str]] = None,
                 instrument_types: Optional[List[str]] = None,
                 data_product_types: Optional[List[str]] = None,
                 metadata_filters: Optional[List[Dict]] = None,
                 metadata_filters_logical_operator: Optional[str] = None,
                 response_format: Optional[Dict] = None) -> Search:
    """
    Submit a request for a data products search, return immediately.

    The request will be done asynchronously by the API. Use the helper functions
    as part of the Search object returned to check for data and/or download it.
    If you don't want the search to return immediately and rather block until
    all data is downloaded, please use the 'search' function instead.

    Note: At least one search criteria from programs, platforms, or
    instrument_types, must be specified.

    .. deprecated::
        This function is deprecated as of 0.9.0. Please use the 'search' function
        with the 'return_immediately' flag set to True to get the same behaviour.
        This function will be removed in a future release.

    Args:
        start: start timestamp of the search (inclusive)
        end: end timestamp of the search (inclusive)
        programs: list of programs to search through, defaults to None
        platforms: list of platforms to search through, defaults to None
        instrument_types: list of instrument types to search through, defaults to None
        data_product_types: list of dictionaries describing data product
            types to filter on e.g. "keogram", defaults to None. Options are in the
            pyaurorax.data_products module, or at the top level using the
            pyaurorax.DATA_PRODUCT_TYPE* variables.
        metadata_filters: list of dictionaries describing metadata keys and
            values to filter on, defaults to None

            Example:

                [{
                    "key": "nbtrace_region",
                    "operator": "in",
                    "values": ["north polar cap"]
                }]
        metadata_filters_logical_operator: the logical operator to use when
            evaluating metadata filters (either 'AND' or 'OR'), defaults
            to "AND"
        response_format: JSON representation of desired data response format

    Returns:
        a pyaurorax.data_products.Search object
    """
    warnings.warn("This function is deprecated and will be removed in a future release. Please "
                  "use the 'search' function with the 'return_immediately' flag to produce the "
                  "same behaviour.", DeprecationWarning, stacklevel=2)
    s = Search(start,
               end,
               programs=programs,
               platforms=platforms,
               instrument_types=instrument_types,
               data_product_types=data_product_types,
               metadata_filters=metadata_filters,
               metadata_filters_logical_operator=metadata_filters_logical_operator,
               response_format=response_format)
    s.execute()
    return s


def upload(identifier: int,
           records: List[DataProduct],
           validate_source: Optional[bool] = False) -> int:
    """
    Upload data product records to AuroraX

    Args:
        identifier: the AuroraX data source ID
        records: data product records to upload
        validate_source: validate all records before uploading, defaults to False

    Returns:
        1 for success, raises exception on error

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected content error
        pyaurorax.exceptions.AuroraXUploadException: upload error
        pyaurorax.exceptions.AuroraXValidationException: data source validation error
    """
    # validate record sources if the flag is set
    if validate_source:
        validation_error = __validate_data_source(identifier, records)
        if validation_error:
            raise AuroraXValidationException(f"Unable to validate data source "
                                             "found in record: {validation_error}")

    # translate each data product record to a request-friendly
    # dict (ie. convert datetimes to strings, etc.)
    for i, _ in enumerate(records):
        if (type(records[i]) is DataProduct):
            records[i] = records[i].to_json_serializable()  # type: ignore

    # make request
    url = api_urls.data_products_upload_url.format(identifier)
    req = AuroraXRequest(method="post",
                         url=url,
                         body=records,
                         null_response=True)
    res = req.execute()

    # evaluate response
    if (res.status_code == 400):
        if type(res.data) is list:
            raise AuroraXUploadException("%s - %s" % (res.status_code,
                                                      res.data[0]["error_message"]))

        raise AuroraXUploadException("%s - %s" % (res.status_code,
                                                  res.data["error_message"]))

    # return
    return 1


def delete_urls(data_source: DataSource,
                urls: List[str]) -> int:
    """
    Delete data products by URL.

    The API processes this request asynchronously, so this method will return
    immediately whether or not the data has already been deleted.

    Args:
        data_source: data source associated with the data product records (note that
            identifier, program, platform, and instrument_type are required)
        urls: URLs of data product records to delete

    Returns:
        1 on success

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXBadParametersException: invalid parameters entered
        pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
    """
    # check to make sure the identifier, program, platform, and instrument type are all set in the data source
    if not all([data_source.identifier, data_source.program, data_source.platform, data_source.instrument_type]):
        raise AuroraXBadParametersException("One or more required data source parameters "
                                            "are missing, delete operation aborted")

    # do request
    url = api_urls.data_products_upload_url.format(data_source.identifier)
    params = {
        "program": data_source.program,
        "platform": data_source.platform,
        "instrument_type": data_source.instrument_type,
        "urls": urls
    }
    delete_req = AuroraXRequest(method="delete",
                                url=url,
                                body=params,
                                null_response=True)
    res = delete_req.execute()

    # evaluate response
    if (res.status_code == 400):
        if type(res.data) is list:
            raise AuroraXBadParametersException("%s - %s" % (res.status_code,
                                                             res.data[0]["message"]))
        raise AuroraXBadParametersException("%s - %s" % (res.status_code,
                                                         res.data["message"]))

    # return
    return 1


def delete(data_source: DataSource,
           start: datetime.datetime,
           end: datetime.datetime,
           data_product_types: Optional[List[str]] = None) -> int:
    """
    Delete data products associated with a data source within a date range.

    The API processes this request asynchronously, so this method will return
    immediately whether or not the data has already been deleted.

    Args:
        data_source: data source associated with the data product records (note that
            identifier, program, platform, and instrument_type are required)
        start: timestamp marking beginning of range to delete records for, inclusive
        end: timestamp marking end of range to delete records for, inclusive
        data_product_types: specific types of data product to delete, e.g.
            ["keogram", "movie"]. If omitted, all data product types will be deleted.

    Returns:
        1 on success

    Raises:
        pyaurorax.exceptions.AuroraXMaxRetriesException: max retry error
        pyaurorax.exceptions.AuroraXUnexpectedContentTypeException: unexpected error
        pyaurorax.exceptions.AuroraXNotFoundException: source not found
        pyaurorax.exceptions.AuroraXUnauthorizedException: invalid API key for this operation
    """
    # check to make sure the identifier, program, platform, and instrument type are all set in the data source
    if not all([data_source.identifier, data_source.program, data_source.platform, data_source.instrument_type]):
        raise AuroraXBadParametersException("One or more required data source parameters "
                                            "are missing, delete operation aborted")

    # do request to get all data products between start and end datetimes
    try:
        s = search(start,
                   end,
                   programs=[data_source.program],
                   platforms=[data_source.platform],
                   instrument_types=[data_source.instrument_type],
                   data_product_types=[] if not data_product_types else data_product_types)
    except Exception as e:
        raise AuroraXException(e)

    # collect URLs from search result
    urls = []
    for dp in s.data:
        urls.append(dp.url)  # type: ignore

    # do delete request
    return delete_urls(data_source, urls)


def describe(search_obj: Search) -> str:
    """
    Describe a data product search as an "SQL-like" string

    Args:
        search_obj: the data product search object to describe

    Returns:
        the "SQL-like" string describing the data product search object
    """
    # make request
    req = AuroraXRequest(method="post",
                         url=api_urls.describe_data_products_query,
                         body=search_obj.query)
    res = req.execute()

    # return
    return res.data


def get_request_url(request_id: str) -> str:
    """
    Get the data product search request URL for a given
    request ID. This URL can be used for subsequent
    pyaurorax.requests function calls. Primarily this method
    facilitates delving into details about a set of already-submitted
    data product searches.

    Args:
        request_id: the request identifier

    Returns:
        the request URL
    """
    url = api_urls.data_products_request_url.format(request_id)
    return url



"""
The availability module provides functions to quickly
determine what data exists on the AuroraX platform.

Note that all functions and classes from submodules are all imported
at this level of the availability module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# function and class imports
from .availability import (ephemeris,
                           data_products)
from .classes.availability_result import AvailabilityResult

# pdoc imports and exports
from .availability import __pdoc__ as __availability_pdoc__
from .classes.availability_result import __pdoc__ as __classes_avail_result_pdoc__
__pdoc__ = __availability_pdoc__
__pdoc__ = dict(__pdoc__, **__classes_avail_result_pdoc__)
__all__ = [
    "ephemeris",
    "data_products",
    "AvailabilityResult",
]



import sys
import os
import click
import pprint
import datetime
import humanize
import textwrap
import warnings
import json
import pyaurorax
from dateutil.parser import parse
from termcolor import colored
from texttable import Texttable


def __echo_helper(message, show_times=False):
    if (show_times is True):
        click.echo("[%s] %s" % (datetime.datetime.now(), message))
    else:
        click.echo(message)


def print_request_logs_table(logs, filter_level=None, table_max_width=None, truncate=True):
    """
    Function to print request logs table

    This is a shared helper function because it is used by the
    conjunction, data products, and ephemeris command modules.
    """
    # init
    default_wrap_threshold = 70
    if (table_max_width is None):
        try:
            terminal_columns = os.get_terminal_size().columns
            wrap_threshold = terminal_columns - 38 - 5  # 38 is the size level+timestamp, 5 for some padding
            if (wrap_threshold < 0):
                warnings.warn("Terminal width is too small, using default "
                              "table width which might not look good")
                wrap_threshold = default_wrap_threshold
        except Exception:
            wrap_threshold = default_wrap_threshold
    else:
        wrap_threshold = table_max_width - 38 - 5  # 38 is the size level+timestamp, 5 for some padding
        if (wrap_threshold < 0):
            warnings.warn("Terminal width is too small, using default "
                          "table width which might not look good")
            wrap_threshold = default_wrap_threshold

    # set table lists
    table_levels = []
    table_summaries = []
    table_timestamps = []
    for log in logs:
        if (filter_level is None or log["level"] == filter_level):
            table_levels.append(log["level"])
            if (truncate is True and len(log["summary"]) > 1000):
                table_summaries.append('\n'.join(textwrap.wrap(log["summary"][0:1000] + "...", wrap_threshold)))
            else:
                table_summaries.append('\n'.join(textwrap.wrap(log["summary"], wrap_threshold)))
            table_timestamps.append(parse(log["timestamp"], ignoretz=True))
    # set header values
    table_headers = ["Timestamp", "Level", "Summary"]

    # output information
    table = Texttable(max_width=400)
    table.set_deco(Texttable.HEADER)
    table.set_cols_dtype(["t"] * len(table_headers))
    table.set_header_align(["l"] * len(table_headers))
    table.set_cols_align(["l"] * len(table_headers))
    table.header(table_headers)
    for i in range(0, len(table_timestamps)):
        table.add_row([table_timestamps[i],
                       table_levels[i],
                       table_summaries[i]])
    click.echo(table.draw())


def print_request_status(s, show_logs=False, show_query=False,
                         filter_logs=None, table_max_width=None):
    """
    Function to print request status information

    This is a shared helper function because it is used by the
    conjunction, data products, and ephemeris command modules.
    """
    # set formatted output variables
    request_completed = colored("False", "yellow")
    request_completed_timestamp = "-"
    request_started_timestamp = parse(s["search_request"]["requested"], ignoretz=True)
    error_condition = "-"
    query_duration = "-"
    data_url = "-"
    file_size = "-"
    result_count = "-"
    if (s["search_result"]["completed_timestamp"] is not None):
        # set completed and completed timestamp
        request_completed = "True"
        request_completed_timestamp = parse(s["search_result"]["completed_timestamp"], ignoretz=True)

        # humanize some values
        query_duration = "%s (%.0fms)" % (humanize.precisedelta(
            datetime.timedelta(milliseconds=s["search_result"]["query_duration"])),
            s["search_result"]["query_duration"])
        data_url = "%s%s" % (pyaurorax.api.get_base_url(), s["search_result"]["data_uri"])
        file_size = humanize.naturalsize(s["search_result"]["file_size"])
        result_count = humanize.intcomma(s["search_result"]["result_count"])

        # set error condition
        error_condition = "False"
        if (s["search_result"]["error_condition"] is True):
            error_condition = colored("True", "red")

    # print out status nicely
    click.echo("Completed:\t\t%s" % (request_completed))
    click.echo("Requested timestamp:\t%s" % (request_started_timestamp))
    click.echo("Completed timestamp:\t%s" % (request_completed_timestamp))
    click.echo("Query duration:\t\t%s" % (query_duration))
    click.echo("Error condition:\t%s" % (error_condition))
    click.echo("Data URL:\t\t%s" % (data_url))
    click.echo("File size:\t\t%s" % (file_size))
    click.echo("Result count:\t\t%s" % (result_count))

    # print out logs if we asked for them
    if (show_logs is True):
        if ("logs" in s):
            click.echo()
            print_request_logs_table(s["logs"],
                                     filter_level=filter_logs,
                                     table_max_width=table_max_width)
        else:
            click.echo("Search logs: missing, unable to display")

    # print out query if we asked for it
    if (show_query is True):
        if ("query" in s["search_request"]):
            click.echo("\nSearch query:\n==================")
            query_to_show = s["search_request"]["query"]
            del query_to_show["request_id"]
            click.echo(pprint.pformat(query_to_show))
        else:
            click.echo("\nSearch query: missing, unable to display")


def get_search_data(type, request_uuid, outfile, output_to_terminal,
                    indent, minify, show_times=False, search_obj=None):
    """
    Function to get search request data

    This is a shared helper function because it is used by the
    conjunction, data products, and ephemeris command modules.
    Some if statements are used to differentiate between the
    commans.
    """
    # get the data
    try:
        # check the status if we need to
        if (search_obj is None):
            # set status url
            __echo_helper("Checking request status ...", show_times=show_times)
            if (type == "conjunctions"):
                url = pyaurorax.api.urls.conjunction_request_url.format(request_uuid)
            elif (type == "data_products"):
                url = pyaurorax.api.urls.data_products_request_url.format(request_uuid)
            elif (type == "ephemeris"):
                url = pyaurorax.api.urls.ephemeris_request_url.format(request_uuid)
            else:
                click.echo("Unexpected error occurred, please open an issue on "
                           "the Github repository detailing how you were able "
                           "to make this message appear")
                sys.exit(1)

            # get status
            try:
                s = pyaurorax.requests.get_status(url)
            except pyaurorax.AuroraXNotFoundException:
                click.echo("Error: request ID not found")
                sys.exit(1)

            # check status
            if (s["search_result"]["completed_timestamp"] is None):
                click.echo("Error: Search is not done yet, not retrieving data")
                click.echo("\nNote: you can use the get_status command to "
                           "check if the search has completed. Try the command "
                           "\"aurorax-cli %s get_status %s\"" % (type, request_uuid))
                sys.exit(1)

            # set data url
            data_url = "%s/data" % (url)
        else:
            s = search_obj.status
            data_url = search_obj.data_url

        # get data
        try:
            __echo_helper("Downloading %s results and %s of data ..." % (humanize.intcomma(s["search_result"]["result_count"]),
                                                                         humanize.naturalsize(s["search_result"]["file_size"])),
                          show_times=show_times)
            data = pyaurorax.requests.get_data(data_url, skip_serializing=True)
        except pyaurorax.AuroraXDataRetrievalError as e:
            # parse error message
            if ("NotFound" in (str(e))):
                click.echo("\n%s" % ('\n'.join(textwrap.wrap("Error downloading data: this request is too old and the data "
                                                             "file has been removed on the server. You can re-run the "
                                                             "search using the command \"aurorax-cli %s "
                                                             "search_resubmit %s\"." % (type, request_uuid), 110))))
            else:
                __echo_helper("Error downloading data: %s" % (str(e)), show_times=show_times)
            sys.exit(1)
    except pyaurorax.AuroraXNotFoundException as e:
        click.echo("%s occurred: request ID not found" % (type(e).__name__))
        sys.exit(1)
    except pyaurorax.AuroraXException as e:
        __echo_helper("%s occurred: %s" % (type(e).__name__, e.args[0]), show_times=show_times)
        sys.exit(1)

    # order data
    if (type == "conjunctions"):
        data = sorted(data, key=lambda x: x["start"])
    elif (type == "data_products"):
        data = sorted(data, key=lambda x: x["start"])
    elif (type == "ephemeris"):
        data = sorted(data, key=lambda x: x["epoch"])

    # save data to file, or print out
    if (output_to_terminal is not None):
        # print out to the terminal (don't save to file)
        #
        # format the data to objects if that was requested
        click.echo()  # one line spacer
        if (output_to_terminal == "dict"):
            # print dict format
            pprint.pprint(data)
        else:
            # serialize the data and print
            for d in data:
                if (type == "conjunctions"):
                    # serialize into Conjunction object
                    d_serialized = pyaurorax.conjunctions.Conjunction(**d, format=pyaurorax.FORMAT_BASIC_INFO)
                elif (type == "data_products"):
                    # serialize into DataProduct object
                    d_serialized = pyaurorax.data_products.DataProduct(**d, format=pyaurorax.FORMAT_BASIC_INFO)
                elif (type == "ephemeris"):
                    # serialize into Ephemeris object
                    d_serialized = pyaurorax.ephemeris.Ephemeris(**d, format=pyaurorax.FORMAT_BASIC_INFO)
                else:
                    click.echo("Unexpected error occurred, please open an issue on "
                               "the Github repository detailing how you were able "
                               "to make this message appear")
                    sys.exit(1)

                # print to terminal
                click.echo(d_serialized)
    else:
        # write to file
        #
        # set filename
        if (outfile is None):
            outfile = "%s_data.json" % (request_uuid)

        # write data to the file
        __echo_helper("Writing data to file ...", show_times=show_times)
        with open(outfile, 'w', encoding="utf-8") as fp:
            if (minify is True):
                json.dump(data, fp)
            else:
                json.dump(data, fp, indent=indent)
        __echo_helper("Data has been saved to '%s'" % (outfile), show_times=show_times)



"""
AuroraX metadata schemas describe the intended structure of metadata stored in
ephemeris and data product records. This module provides functions for
interacting with the schemas.

Note that all functions and classes from submodules are all imported
at this level of the metadata module. They can be referenced from
here instead of digging in deeper to the submodules.
"""

# function and class imports
from .metadata import (get_data_products_schema,
                       get_ephemeris_schema,
                       validate)

# pdoc imports and exports
from .metadata import __pdoc__ as __metadata_pdoc__
__pdoc__ = __metadata_pdoc__
__all__ = [
    "get_data_products_schema",
    "get_ephemeris_schema",
    "validate",
]



"""
The Location module provides a class used throughout the PyAuroraX
library to manage lat/lon positions of different things.
"""

from pydantic import BaseModel, validator
from typing import Union, Optional


class Location(BaseModel):
    """
    Class representing an AuroraX location (ie. geographic coordinates,
    GSM coordinates, northern/southern B-trace magnetic footprints)

    The numbers are in decimal degrees format and range from -90 to 90
    for latitude and -180 to 180 for longitude.

    Attributes:
        lat: latitude value
        lon: longitude value
    """
    lat: Optional[Union[float, None]] = None
    lon: Optional[Union[float, None]] = None

    @validator("lon")
    def __both_must_be_none_or_number(cls, v, values):  # pylint: disable=unused-private-member
        # check to make sure the values are both numbers or None types. We don't
        # allow a Location object to have the latitude set and not the
        # longitude (or vice-versa)
        if (v and not values["lat"]) or (values["lat"] and not v):
            raise ValueError("Latitude and longitude must both be numbers, or both be None")
        return v

    def __str__(self) -> str:
        """
        String method

        Returns:
            string format of Location object
        """
        return self.__repr__()

    def __repr__(self) -> str:
        """
        Object representation

        Returns:
            object representation of Location object
        """
        return "%s(lat=%s, lon=%s)" % (self.__class__.__name__, str(self.lat), str(self.lon))



"""
The PyAuroraX package provides a way to interact with the
[AuroraX API](https://aurorax.space/data/apiLibraries). It is intended
to provide an intuitive process for those in the space physics and related
communities to programmatically query AuroraX's vast database for conjunctions,
ephemeris or data product records, data availability information, and more.
PyAuroraX requires Python 3.6, 3.7, 3.8, or 3.9 (Python 3.10 currently not
supported).

Check out this project on [GitHub](https://github.com/aurorax-space/pyaurorax)
and explore the evolving ecosystem of visualizations, tools, and data
at [AuroraX](https://aurorax.space/).

For an overview of usage and examples, visit the
[AuroraX Documentation website](https://docs.aurorax.space/code/overview).
Details of functionality and options are available in the
[API reference](https://docs.aurorax.space/code/pyaurorax_api_reference/pyaurorax/).

Installation:
```console
$ python -m pip install pyaurorax
```

Basic usage:
```python
> import pyaurorax
```
"""

# versioning info
__version__ = "0.10.0"

# documentation excludes
__pdoc__ = {"cli": False}

# pull in top level functions
from .api import (AuroraXRequest,
                  authenticate,
                  get_api_key)
from .sources import (FORMAT_BASIC_INFO,
                      FORMAT_BASIC_INFO_WITH_METADATA,
                      FORMAT_FULL_RECORD,
                      FORMAT_IDENTIFIER_ONLY,
                      FORMAT_DEFAULT,
                      SOURCE_TYPE_EVENT_LIST,
                      SOURCE_TYPE_GROUND,
                      SOURCE_TYPE_HEO,
                      SOURCE_TYPE_LEO,
                      SOURCE_TYPE_LUNAR)
from .conjunctions import (CONJUNCTION_TYPE_NBTRACE,
                           CONJUNCTION_TYPE_SBTRACE)
from .data_products import (DATA_PRODUCT_TYPE_KEOGRAM,
                            DATA_PRODUCT_TYPE_MONTAGE,
                            DATA_PRODUCT_TYPE_MOVIE,
                            DATA_PRODUCT_TYPE_SUMMARY_PLOT,
                            DATA_PRODUCT_TYPE_DATA_AVAILABILITY)

# pull in exceptions at top level
from .exceptions import (AuroraXException,
                         AuroraXNotFoundException,
                         AuroraXMaxRetriesException,
                         AuroraXDuplicateException,
                         AuroraXUnexpectedContentTypeException,
                         AuroraXValidationException,
                         AuroraXBadParametersException,
                         AuroraXUnauthorizedException,
                         AuroraXConflictException,
                         AuroraXUploadException,
                         AuroraXUnexpectedEmptyResponse,
                         AuroraXDataRetrievalError)

# pull in models
from .location import Location

# pull in modules (order matters otherwise we get circular import errors)
from pyaurorax import requests
from pyaurorax import api
from pyaurorax import sources
from pyaurorax import exceptions
from pyaurorax import metadata
from pyaurorax import util
from pyaurorax import availability
from pyaurorax import conjunctions
from pyaurorax import ephemeris
from pyaurorax import data_products
